from __future__ import annotations

import json
import logging
import os

import h5py

import pygama.lgdo as lgdo
from pygama.lgdo import LH5Store
from pygama.raw.buffer_processor.buffer_processor import buffer_processor
from pygama.raw.raw_buffer import RawBuffer, RawBufferLibrary

log = logging.getLogger(__name__)


def lh5_buffer_processor(
    lh5_raw_file_in: str,
    overwrite: bool = False,
    out_spec: dict = None,
    proc_file_name: str = None,
) -> None:
    r"""Process raw buffers from an LH5 file.

    This function performs data processing on an existing raw-level LH5 file
    according to `out_spec`.  It iterates through any valid table in the
    `lh5_raw_file_in` file and checks that either the `group_name` or the
    `out_name` of this table are found in the `out_spec`. If there is a valid
    `proc_spec` entry in the `out_spec`, then the table is stored in a
    :class:`RawBuffer` and the :func:`RawBuffer` is passed to the buffer
    processor. If no `proc_spec` is found, the table is written to the file
    without any modifications.


    Parameters
    ----------
    lh5_raw_file_in
        the path of a file created by :func:`~.raw.build_raw.build_raw`.
    overwrite
        sets whether to overwrite the output file(s) if it (they) already
        exist.
    out_spec
        an `out_spec` for :func:`~.raw.build_raw.build_raw` also containing a
        dictionary named `proc_spec` used for data processing.  See
        :func:`.buffer_processor` for an example of such an `out_spec`.
    proc_file_name
        The path to the processed output file created. If ``None``, uses
        ``{lh5_raw_file_in}_proc.lh5`` as the output filename.

    Notes
    -----
    This function assumes that all raw file data are stored in at most two
    :class:`h5py.Group`\ s, as in e.g. ``ch000/raw`` or ``raw/geds``
    """

    # Initialize the input raw file
    raw_store = LH5Store()
    lh5_file = raw_store.gimme_file(lh5_raw_file_in, "r")
    if lh5_file is None:
        raise ValueError(f"input file not found: {lh5_raw_file_in}")
        return

    # List the groups in the raw file
    lh5_groups = lgdo.ls(lh5_raw_file_in)
    lh5_tables = []

    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for tb in lh5_groups:
        # Make sure that the upper level key isn't a dataset
        if isinstance(lh5_file[tb], h5py.Dataset):
            lh5_tables.append(f"{tb}")
        elif "raw" not in tb and lgdo.ls(lh5_file, f"{tb}/raw"):
            lh5_tables.append(f"{tb}/raw")
        # Look one layer deeper for a :meth:`lgdo.Table` if necessary
        elif lgdo.ls(lh5_file, f"{tb}"):
            # Check to make sure that this isn't a table itself
            maybe_table, _ = raw_store.read_object(f"{tb}", lh5_file)
            if isinstance(maybe_table, lgdo.Table):
                lh5_tables.append(f"{tb}")
                del maybe_table
            # otherwise, go deeper
            else:
                for sub_table in lgdo.ls(lh5_file, f"{tb}"):
                    maybe_table, _ = raw_store.read_object(
                        f"{tb}/{sub_table}", lh5_file
                    )
                    if isinstance(maybe_table, lgdo.Table):
                        lh5_tables.append(f"{tb}/{sub_table}")
                    del maybe_table

    if len(lh5_tables) == 0:
        raise RuntimeError(f"could not find any valid LH5 table in {lh5_raw_file_in}")

    # Initialize the processed file
    # Get the proc_file_name if the user doesn't pass one
    if proc_file_name is None:
        proc_file_name = lh5_raw_file_in
        i_ext = proc_file_name.rfind(".")
        if i_ext != -1:
            proc_file_name = proc_file_name[:i_ext]
        proc_file_name += "_proc.lh5"
    # clear existing output files
    if overwrite:
        if os.path.isfile(proc_file_name):
            os.remove(proc_file_name)
    raw_store.gimme_file(proc_file_name, "a")

    # Do key expansion on the out_spec
    if isinstance(out_spec, str) and out_spec.endswith(".json"):
        with open(out_spec) as json_file:
            out_spec = json.load(json_file)
    if isinstance(out_spec, dict):
        RawBufferLibrary(json_dict=out_spec)

    # Write everything in the raw file to the new file, check for proc_spec under either the group name, out_name, or the name
    for tb in lh5_tables:
        lgdo_obj, _ = raw_store.read_object(f"{tb}", lh5_file)

        # Find the out_name.
        # If the top level group has an lgdo table in it, then the out_name is group
        if len(tb.split("/")) == 1:
            out_name = tb.split("/")[0]
            group_name = "/"  # There is no group name
        # We are only considering nested groups of the form key1/key2/, so key2 is the out name
        else:
            out_name = tb.split("/")[-1]  # the out_name is the last key
            group_name = tb.split("/")[0]  # the group name is the first key

        for decoder_name in out_spec.keys():
            # Check to see if the group name is in the out_spec's dict
            # This is used if there is key expansion
            if group_name in out_spec[decoder_name].keys():
                # Check to see if that group in the out_spec has a `proc_spec` key:
                if "proc_spec" in out_spec[decoder_name][group_name].keys():
                    # Recast the lgdo_obj as a RawBuffer
                    rb = RawBuffer(
                        lgdo=lgdo_obj,
                        out_name=out_name,
                        proc_spec=out_spec[decoder_name][group_name]["proc_spec"],
                    )
                    tmp_table = buffer_processor(rb)
                    # Update the lgdo_obj to be written to the processed file
                    lgdo_obj = tmp_table
                else:
                    pass

            # If there is no key expansion, and no out_name in the out_spec, then the out_name could be in the out_spec's keys, i.e. the name
            elif out_name in out_spec[decoder_name].keys():
                # Check to see if that group in the out_spec has a `proc_spec` key:
                if "proc_spec" in out_spec[decoder_name][out_name].keys():
                    # Recast the lgdo_obj as a RawBuffer
                    rb = RawBuffer(
                        lgdo=lgdo_obj,
                        out_name=out_name,
                        proc_spec=out_spec[decoder_name][out_name]["proc_spec"],
                    )
                    tmp_table = buffer_processor(rb)
                    # Update the lgdo_obj to be written to the processed file
                    lgdo_obj = tmp_table
                else:
                    pass

            # Lastly, having an out_name specified in the out_spec could override the name key. Check for it
            elif (
                (len(list(out_spec[decoder_name].keys())) == 1)
                and (
                    "out_name"
                    in out_spec[decoder_name][list(out_spec[decoder_name].keys())[0]]
                )
                and (
                    out_name
                    == out_spec[decoder_name][list(out_spec[decoder_name].keys())[0]][
                        "out_name"
                    ]
                )
            ):
                # if out_name is a key, check that the out_name matches and use that
                if (
                    "proc_spec"
                    in out_spec[decoder_name][
                        list(out_spec[decoder_name].keys())[0]
                    ].keys()
                ):
                    # Recast the lgdo_obj as a RawBuffer
                    rb = RawBuffer(
                        lgdo=lgdo_obj,
                        out_name=out_name,
                        proc_spec=out_spec[decoder_name][
                            list(out_spec[decoder_name].keys())[0]
                        ]["proc_spec"],
                    )
                    tmp_table = buffer_processor(rb)
                    # Update the lgdo_obj to be written to the processed file
                    lgdo_obj = tmp_table
                else:
                    pass
            else:
                pass

        # Write the (possibly processed) lgdo_obj to a file
        raw_store.write_object(
            lgdo_obj, out_name, lh5_file=proc_file_name, group=group_name
        )



from __future__ import annotations

import numpy as np
from numba import guvectorize

import pygama.lgdo.lh5_store as lh5
from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def wiener_filter(file_name_array: list[str]) -> np.ndarray:
    """Apply a Wiener filter to the waveform.

    Note
    ----
    The convolution is performed in the frequency domain.  This processor is
    composed of a factory function that is called using the `init_args`
    argument. The input and output waveforms are passed using `args`. The input
    must be the Fourier transform of the waveform. The output is the filtered
    waveform in the frequency domain.

    Parameters
    ----------
    file_name_array
        Array with path to an HDF5 file containing the time domain version of
        the superpulse in one column and noise waveform in another, the
        superpulse HDF5 group must be titled ``spms/processed/superpulse`` and
        the noise waveform must be called ``spms/processed/noise_wf``.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_wiener": {
            "function": "wiener_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_bl_fft", "wf_wiener(2000,f)"],
            "unit": "dB",
            "init_args": ["/path/to/file/wiener.lh5"]
        }
    """

    sto = lh5.LH5Store()

    # Check that the file is valid and the data is in the correct format

    try:
        file_name_array[0]
    except Exception:
        raise DSPFatal("init_args must be an array with the filename")

    file_name = file_name_array[0]

    try:
        f = sto.gimme_file(file_name, "r")
    except Exception:
        raise DSPFatal("File must be a valid lh5 file")

    if "spms/processed/superpulse" not in f:
        raise DSPFatal("lh5 file must have 'spms/processed/superpulse' as a group")

    if "spms/processed/noise_wf" not in f:
        raise DSPFatal("lh5 file must have 'spms/processed/noise_wf' as a group")

    # Read in the data

    superpulse, _ = sto.read_object("spms/processed/superpulse", file_name)
    superpulse = superpulse.nda

    noise_wf, _ = sto.read_object("spms/processed/noise_wf", file_name)
    noise_wf = noise_wf.nda

    # Now check that the data are valid

    if len(superpulse) <= 0:
        raise DSPFatal("The length of the filter must be positive")

    if len(superpulse) != len(noise_wf):
        raise DSPFatal(
            "The length of the superpulse must be equal to the length of the noise waveform"
        )

    if np.argmax(superpulse) <= 0 or np.argmax(superpulse) > len(superpulse):
        raise DSPFatal(
            "The index of the maximum of the superpulse must occur within the waveform"
        )

    # Transform these to the frequency domain to eventually create the wiener filter

    fft_superpulse = np.fft.fft(superpulse)
    fft_noise_wf = np.fft.fft(noise_wf)

    # Create the point spread function for the detector's response

    def psf(superpulse, fft_superpulse):
        delta = np.zeros_like(superpulse)
        arg_max = np.argmax(superpulse)
        delta[arg_max] = np.amax(superpulse)

        return fft_superpulse / np.fft.fft(delta)

    # Now create the wiener filter in the frequency domain

    fft_psf = psf(superpulse, fft_superpulse)
    psd_noise_wf = fft_noise_wf * np.conj(fft_noise_wf)
    psd_superpulse = fft_superpulse * np.conj(fft_superpulse)

    w_filter = (np.conj(fft_psf)) / (
        (fft_psf * np.conj(fft_psf)) + (psd_noise_wf / psd_superpulse)
    )

    # Create a factory function that performs the convolution with the wiener filter, the output is still in the frequency domain

    @guvectorize(
        ["void(complex64[:], complex64[:])", "void(complex128[:], complex128[:])"],
        "(n)->(n)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def wiener_out(fft_w_in: np.ndarray, fft_w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        fft_w_in
            the Fourier transformed input waveform.
        fft_w_out
            the filtered waveform, in the frequency domain.
        """
        fft_w_out[:] = np.nan

        if np.isnan(fft_w_in).any():
            return

        if len(w_filter) != len(fft_w_in):
            raise DSPFatal("The filter is not the same length of the input waveform")

        fft_w_out[:] = fft_w_in * w_filter

    return wiener_out



import json
import plistlib
import re
import struct
from collections import Counter
from io import BytesIO

from pygama.lgdo import LH5Store
from pygama.raw import build_raw
from pygama.raw.orca import orca_streamer
from pygama.raw.orca.orca_flashcam import ORFlashCamListenerConfigDecoder
from pygama.raw.orca.orca_run_decoder import ORRunDecoderForRun


class OrcaEncoder:
    """Encoding function for recreating raw Orca files."""

    def __init__(self, file):
        """Initialize the encoder for daq to raw testing."""

        self.file = file
        self.header = None

    def encode_header(self):
        """Convert orca header back to a byte string."""

        lh5 = LH5Store()
        test_file, _ = lh5.read_object(
            "OrcaHeader",
            self.file,
        )
        test_str = str(test_file.value, "utf-8")
        xml_dict = json.loads(test_str)
        xml_header = BytesIO()
        plistlib.dump(xml_dict, xml_header, fmt=plistlib.FMT_XML)
        xml_header.seek(0)
        compiled_header = str(xml_header.read(), "utf-8")
        xml_header.close()
        fixed_header = re.sub(
            r"([1-9]\d*\.0)+",
            lambda x: str(int(float(x.group(0)))),
            compiled_header,
        ).replace("1", "1.0", 4)

        self.header = bytes(fixed_header, "utf-8")
        len_header_true = len(self.header)

        if (len_header_true + 8) % 16 != 0:
            extra_zeros = 16 - ((len_header_true + 8) % 16)
            extra_bytes = struct.pack(
                f"{extra_zeros}b",
                *([0] * extra_zeros),
            )
            self.header += extra_bytes

        len_header_full = len(self.header) + 8

        init_header = struct.pack("i", (len_header_full) // 4)
        init_header += struct.pack("i", len_header_true)

        self.header = init_header + self.header

    def encode_orflashcamconfig(self, ii):
        """Convert orca flashcam config data back to byte strings."""

        lh5 = LH5Store()
        tbl, _ = lh5.read_object(
            "ORFlashCamListenerConfig",
            self.file,
        )

        packets = []
        packets.append(4 << 18)
        packets.append((tbl["readout_id"].nda[ii] << 16) + tbl["fcid"].nda[ii])

        decoded_values = ORFlashCamListenerConfigDecoder().get_decoded_values()

        for i, k in enumerate(decoded_values):
            if i < 4:
                continue
            packets.append(tbl[k].nda[ii])
            if k == "gps":
                break

        bvi0 = 0  # start index of board vector-of-vector's
        if ii > 0:
            bvi0 = tbl["ch_board_id"].cumulative_length.nda[ii - 1]
        npacks = tbl["ch_board_id"].cumulative_length.nda[ii] - bvi0

        for jj in range(npacks):
            board_id = tbl["ch_board_id"].flattened_data.nda[bvi0 + jj]
            fc_input = tbl["ch_inputnum"].flattened_data.nda[bvi0 + jj]
            packets.append((board_id << 16) + fc_input)

        while len(packets) < tbl["packet_len"].nda[ii]:
            packets.append(0)

        packets[0] += len(packets)

        return packets

    def encode_orflashcamadcwaveform(self, ii):
        """Convert orca flashcam ADC waveform data back to byte strings."""

        lh5 = LH5Store()
        tbl, _ = lh5.read_object(
            "ORFlashCamADCWaveform",
            self.file,
        )

        orca_header_length = 3
        fcio_header_length = 17

        packets = []
        packets.append(3 << 18)
        packets.append(1)
        packets[1] += orca_header_length << 28
        packets[1] += fcio_header_length << 22

        packet3 = 0
        packet3 += tbl["channel"].nda[ii]
        packet3 += tbl["fc_input"].nda[ii] << 10
        packet3 += (tbl["board_id"].nda[ii] & 0xFF) << 14  # old bad board_id encoding
        packet3 += tbl["slot"].nda[ii] << 22
        packet3 += tbl["crate"].nda[ii] << 27
        packets.append(packet3)

        # time offsets
        packets.append(tbl["mu_offset_sec"].nda[ii])
        packets.append(tbl["mu_offset_usec"].nda[ii])
        packets.append(tbl["to_master_sec"].nda[ii])
        packets.append(tbl["delta_mu_usec"].nda[ii])
        packets.append(tbl["abs_delta_mu_usec"].nda[ii])
        packets.append(tbl["to_start_sec"].nda[ii])
        packets.append(tbl["to_start_usec"].nda[ii])

        # set the dead region values
        packets.append(tbl["dr_start_pps"].nda[ii])
        packets.append(tbl["dr_start_ticks"].nda[ii])
        packets.append(tbl["dr_stop_pps"].nda[ii])
        packets.append(tbl["dr_stop_ticks"].nda[ii])
        packets.append(tbl["dr_maxticks"].nda[ii])

        # set event number and clock counters
        packets.append(tbl["eventnumber"].nda[ii])
        packets.append(tbl["ts_pps"].nda[ii])
        packets.append(tbl["ts_ticks"].nda[ii])
        packets.append(tbl["ts_maxticks"].nda[ii])

        packets.append(tbl["baseline"].nda[ii] + (tbl["daqenergy"].nda[ii] << 16))

        packets.extend(
            [
                xx + (yy << 16)
                for xx, yy in zip(
                    tbl["waveform"]["values"].nda[ii, ::2],
                    tbl["waveform"]["values"].nda[ii, 1::2],
                )
            ]
        )

        wf_samples = 2 * (len(packets) - orca_header_length - fcio_header_length)

        packets[1] += wf_samples << 6
        packets[0] += len(packets)

        return packets

    def encode_orrun(self, ii):
        """Convert orca run data back to byte strings."""

        lh5 = LH5Store()
        tbl, _ = lh5.read_object(
            "ORRunDecoderForRun",
            self.file,
        )

        packets = []
        packets.append(7 << 18)
        packets.append(tbl["subrun_number"].nda[ii] << 16)
        decoded_values = ORRunDecoderForRun().get_decoded_values()
        for i, k in enumerate(decoded_values):
            if 0 < i < 7:
                packets[1] += tbl[k].nda[ii] << (i - 1)

        packets.append(tbl["run_number"].nda[ii])
        packets.append(tbl["time"].nda[ii])

        packets[0] += len(packets)

        return packets


def test_daq_to_raw(lgnd_test_data):
    """Test function for the daq to raw validation."""

    # open orca daq file and create LH5 file
    orca_file = lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca")
    out_spec = "/tmp/L200-comm-20220519-phy-geds_test.lh5"

    build_raw(
        orca_file,
        in_stream_type="ORCA",
        out_spec=out_spec,
        overwrite=True,
    )

    # recreate the header
    encoder = OrcaEncoder(out_spec)
    encoder.encode_header()

    rebuilt_orca_data = encoder.header

    # get the order of all of the data IDs
    orstr = orca_streamer.OrcaStreamer()
    orstr.open_stream(orca_file)
    raw_packets = []

    while True:
        this_packet = orstr.load_packet(skip_unknown_ids=False)
        if this_packet is None:
            break
        raw_packets.append(this_packet.copy())

    orstr.close_stream()

    data_ids = [pack[0] >> 18 for pack in raw_packets]

    # count the number of each ID
    full_count = Counter(data_ids)
    buffer_count = {k: 0 for k in full_count}

    # build the raw orca file in the order of data IDs
    for dd in data_ids:
        buffer_count[dd] += 1
        ii = buffer_count[dd] - 1

        if dd == 7:
            this_packet = encoder.encode_orrun(ii)
        elif dd == 4:
            this_packet = encoder.encode_orflashcamconfig(ii)
        elif dd == 3:
            this_packet = encoder.encode_orflashcamadcwaveform(ii)
        else:
            raise ValueError(f"Encoder does not exist for data ID {dd}")

        rebuilt_orca_data += struct.pack(f"{len(this_packet)}I", *this_packet)

    # load the original raw orca file as a byte string
    with open(orca_file, "rb") as ff:
        orig_orca_data = ff.read()

    # assert the byte strings are the same
    assert len(rebuilt_orca_data) == len(orig_orca_data)
    assert rebuilt_orca_data == orig_orca_data



import numpy as np

from pygama.lgdo import Scalar


def test_decoding(compass_config):
    pass


def test_decoding_no_settings(compass_config_no_settings):
    pass


def test_data_types(compass_config):
    assert isinstance(compass_config, dict)


def test_data_types_no_settings(compass_config_no_settings):
    assert isinstance(compass_config_no_settings, dict)


def test_values(compass_config):
    assert compass_config
    expected_dict = {
        "boards": {
            "0": {
                "adcBitCount": "14",
                "channels": {
                    "0": {
                        "SRV_PARAM_CH_BLINE_DCOFFSET": "20.0",
                        "SRV_PARAM_CH_ENABLED": "true",
                        "SRV_PARAM_CH_INDYN": "INDYN_2_0_VPP",
                        "SRV_PARAM_CH_THRESHOLD": "100.0",
                        "SRV_PARAM_CH_TRG_HOLDOFF": "1024.0",
                    },
                    "1": {
                        "SRV_PARAM_CH_ENABLED": "true",
                        "SRV_PARAM_CH_INDYN": "INDYN_0_5_VPP",
                    },
                },
                "modelName": "DT5730",
                "sample_rate": 0.5,
                "wf_len": 1000.0,
            }
        },
        "energy_calibrated": 0,
        "energy_channels": 1,
        "energy_short": 1,
        "waveform_samples": 1,
    }
    # We have a nested struct, so we need some nasty recursion
    for key in compass_config.keys():
        if type(compass_config[key]) == Scalar:
            assert compass_config[key].value == expected_dict[key]
        # Nested struct
        else:
            for key_2 in compass_config[key]:
                if type(compass_config[key][key_2]) == Scalar:
                    assert compass_config[key][key_2].value == expected_dict[key][key_2]
                else:
                    for key_3 in compass_config[key][key_2]:
                        if type(compass_config[key][key_2][key_3]) == Scalar:
                            assert (
                                compass_config[key][key_2][key_3].value
                                == expected_dict[key][key_2][key_3]
                            )
                        else:
                            for key_4 in compass_config[key][key_2][key_3]:
                                if (
                                    type(compass_config[key][key_2][key_3][key_4])
                                    == Scalar
                                ):
                                    assert (
                                        compass_config[key][key_2][key_3][key_4].value
                                        == expected_dict[key][key_2][key_3][key_4]
                                    )
                                else:
                                    for key_5 in compass_config[key][key_2][key_3][
                                        key_4
                                    ]:
                                        if (
                                            type(
                                                compass_config[key][key_2][key_3][
                                                    key_4
                                                ][key_5]
                                            )
                                            == Scalar
                                        ):
                                            assert (
                                                compass_config[key][key_2][key_3][
                                                    key_4
                                                ][key_5].value
                                                == expected_dict[key][key_2][key_3][
                                                    key_4
                                                ][key_5]
                                            )
                                        else:
                                            raise AssertionError(
                                                "Got more keys than expected"
                                            )


def test_values_no_config(compass_config_no_settings):
    expected_dict = {
        "boards": {
            "0": {
                "adcBitCount": np.nan,
                "channels": {
                    "0": np.nan,
                    "1": np.nan,
                    "2": np.nan,
                    "3": np.nan,
                    "4": np.nan,
                    "5": np.nan,
                    "6": np.nan,
                    "7": np.nan,
                },
                "modelName": np.nan,
                "sample_rate": np.nan,
                "wf_len": 1000.0,
            },
            "1": {
                "adcBitCount": np.nan,
                "channels": {
                    "0": np.nan,
                    "1": np.nan,
                    "2": np.nan,
                    "3": np.nan,
                    "4": np.nan,
                    "5": np.nan,
                    "6": np.nan,
                    "7": np.nan,
                },
                "modelName": np.nan,
                "sample_rate": np.nan,
                "wf_len": 1000.0,
            },
        },
        "energy_calibrated": 0,
        "energy_channels": 1,
        "energy_short": 1,
        "waveform_samples": 1,
    }
    # We have a nested struct, so we need some nasty recursion
    for key in compass_config_no_settings.keys():
        if type(compass_config_no_settings[key]) == Scalar:
            if (np.isnan(compass_config_no_settings[key].value)) and (
                np.isnan(expected_dict[key])
            ):
                assert True
            else:
                assert compass_config_no_settings[key].value == expected_dict[key]
        else:
            for key_2 in compass_config_no_settings[key]:
                if type(compass_config_no_settings[key][key_2]) == Scalar:
                    if (np.isnan(compass_config_no_settings[key][key_2].value)) and (
                        np.isnan(expected_dict[key][key_2])
                    ):
                        assert True
                    else:
                        assert (
                            compass_config_no_settings[key][key_2].value
                            == expected_dict[key][key_2]
                        )
                else:
                    for key_3 in compass_config_no_settings[key][key_2]:
                        if (
                            type(compass_config_no_settings[key][key_2][key_3])
                            == Scalar
                        ):
                            if (
                                np.isnan(
                                    compass_config_no_settings[key][key_2][key_3].value
                                )
                            ) and (np.isnan(expected_dict[key][key_2][key_3])):
                                assert True
                            else:
                                assert (
                                    compass_config_no_settings[key][key_2][key_3].value
                                    == expected_dict[key][key_2][key_3]
                                )
                        else:
                            for key_4 in compass_config_no_settings[key][key_2][key_3]:
                                if (
                                    type(
                                        compass_config_no_settings[key][key_2][key_3][
                                            key_4
                                        ]
                                    )
                                    == Scalar
                                ):
                                    if np.isnan(
                                        compass_config_no_settings[key][key_2][key_3][
                                            key_4
                                        ].value
                                    ) and (
                                        np.isnan(
                                            expected_dict[key][key_2][key_3][key_4]
                                        )
                                    ):
                                        assert True
                                    else:
                                        assert (
                                            compass_config_no_settings[key][key_2][
                                                key_3
                                            ][key_4].value
                                            == expected_dict[key][key_2][key_3][key_4]
                                        )
                                else:
                                    for key_5 in compass_config_no_settings[key][key_2][
                                        key_3
                                    ][key_4]:
                                        if (
                                            type(
                                                compass_config_no_settings[key][key_2][
                                                    key_3
                                                ][key_4][key_5]
                                            )
                                            == Scalar
                                        ):
                                            if (
                                                np.isnan(
                                                    compass_config_no_settings[key][
                                                        key_2
                                                    ][key_3][key_4][key_5].value
                                                )
                                            ) and (
                                                np.isnan(
                                                    expected_dict[key][key_2][key_3][
                                                        key_4
                                                    ][key_5]
                                                )
                                            ):
                                                assert True
                                            else:
                                                assert (
                                                    compass_config_no_settings[key][
                                                        key_2
                                                    ][key_3][key_4][key_5].value
                                                    == expected_dict[key][key_2][key_3][
                                                        key_4
                                                    ][key_5]
                                                )
                                        else:
                                            raise AssertionError(
                                                "Got more keys than expected"
                                            )



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32, float32[:], float32[:], float32[:], float32[:])",
        "void(float64[:], float64, float64, float64, float64, float64, float64[:], float64[:], float64[:], float64[:])",
    ],
    "(n),(),(),(),(),(),(m),(m),(),()",
    **nb_kwargs,
)
def get_multi_local_extrema(
    w_in: np.ndarray,
    a_delta_max_in: float,
    a_delta_min_in: float,
    search_direction: int,
    a_abs_max_in: float,
    a_abs_min_in: float,
    vt_max_out: np.ndarray,
    vt_min_out: np.ndarray,
    n_max_out: int,
    n_min_out: int,
) -> None:
    """Get lists of indices of the local maxima and minima of data.

    Converted from a `MATLAB script <http://billauer.co.il/peakdet.html>`_ by
    E. Billauer. See the parameters description for details about search modes
    and further settings.

    Parameters
    ----------
    w_in
        the array of data within which extrema will be found.
    a_delta_min_in
        the relative level by which data must vary (in the search direction)
        about a candidate minimum in order for it to be tagged.
    a_delta_max_in
        the relative level by which data must vary (in the search direction)
        about a candidate maximum in order for it to be tagged.
    search_direction
        the direction in which the input waveform is processed.

        * ``0``: one sweep, first to last sample
        * ``1``: one sweep, last to first sample
        * ``2``: two sweeps, in both directions. Largest common set of found
          extrema is returned (logical AND)
        * ``3``: two sweeps, in both directions. Union of found extrema is
          returned (logical OR)

    a_abs_min_in
        a candidate minimum is tagged only if its value is smaller than this.
    a_abs_max_in
        a candidate maximum is tagged only if its value is larger than this.
    vt_min_out
        arrays of fixed length (padded with :any:`numpy.nan`) that hold the
        indices of the identified minima.
    vt_max_out
        arrays of fixed length (padded with :any:`numpy.nan`) that hold the
        indices of the identified maxima.
    n_min_out
        the number of minima found in a waveform.
    n_max_out
        the number of maxima found in a waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "vt_max_out, vt_min_out, n_max_out, n_min_out": {
            "function": "get_multi_local_extrema",
            "module": "pygama.dsp.processors",
            "args": [
                "waveform",
                5, 0.1,
                1,
                20, 0,
                "vt_max_out(20)",
                "vt_min_out(20)",
                "n_max_out",
                "n_min_out"
            ],
            "unit": ["ns", "ns", "none", "none"]
        }
    """
    # prepare output
    vt_max_out[:] = np.nan
    vt_min_out[:] = np.nan
    n_max_out[0] = np.nan
    n_min_out[0] = np.nan

    # initialize internal counters
    n_max_left_counter = 0
    n_max_right_counter = 0
    n_min_left_counter = 0
    n_min_right_counter = 0

    # initialize temporary arrays
    left_vt_max = np.zeros(len(vt_max_out), dtype=np.float_)
    left_vt_min = np.zeros(len(vt_min_out), dtype=np.float_)
    right_vt_max = np.zeros(len(vt_max_out), dtype=np.float_)
    right_vt_min = np.zeros(len(vt_min_out), dtype=np.float_)

    left_vt_max[:] = np.nan
    left_vt_min[:] = np.nan
    right_vt_max[:] = np.nan
    right_vt_min[:] = np.nan

    # Checks

    if np.isnan(w_in).any() or np.isnan(a_delta_max_in) or np.isnan(a_delta_min_in):
        return

    if (not len(vt_max_out) < len(w_in)) or (not len(vt_min_out) < len(w_in)):
        raise DSPFatal(
            "The length of your return array must be smaller than the length of your waveform"
        )
    if (not a_delta_max_in >= 0) or (not a_delta_min_in >= 0):
        raise DSPFatal("Delta must be positive")

    # now loop over data
    # left to right search
    if (search_direction == 0) or (search_direction > 1):
        find_max = True
        imax, imin = 0, 0
        for i in range(len(w_in)):
            if w_in[i] > w_in[imax]:
                imax = i
            if w_in[i] < w_in[imin]:
                imin = i
            if find_max:
                # if the sample is less than the current max by more than a_delta_in,
                # declare the previous one a maximum, then set this as the new "min"
                if (
                    (w_in[i] < w_in[imax] - a_delta_max_in)
                    and (int(n_max_left_counter) < int(len(left_vt_max)))
                    and (w_in[imax] > a_abs_max_in)
                ):
                    left_vt_max[int(n_max_left_counter)] = imax
                    n_max_left_counter += 1
                    imin = i
                    find_max = False
            else:
                # if the sample is more than the current min by more than a_delta_in,
                # declare the previous one a minimum, then set this as the new "max"
                if (
                    (w_in[i] > w_in[imin] + a_delta_min_in)
                    and (int(n_min_left_counter) < int(len(left_vt_min)))
                    and (w_in[imin] < a_abs_min_in)
                ):
                    left_vt_min[int(n_min_left_counter)] = imin
                    n_min_left_counter += 1
                    imax = i
                    find_max = True

    # right to left search
    if search_direction > 0:
        find_max = True
        imax, imin = len(w_in) - 1, len(w_in) - 1
        for i in range(len(w_in) - 1, -1, -1):
            if w_in[i] > w_in[imax]:
                imax = i
            if w_in[i] < w_in[imin]:
                imin = i
            if find_max:
                # if the sample is less than the current max by more than a_delta_in,
                # declare the previous one a maximum, then set this as the new "min"
                if (
                    w_in[i] < w_in[imax] - a_delta_max_in
                    and int(n_max_right_counter) < int(len(right_vt_max))
                    and w_in[imax] > a_abs_max_in
                ):
                    right_vt_max[int(n_max_right_counter)] = imax
                    n_max_right_counter += 1
                    imin = i
                    find_max = False
            else:
                # if the sample is more than the current min by more than a_delta_in,
                # declare the previous one a minimum, then set this as the new "max"
                if (
                    w_in[i] > w_in[imin] + a_delta_min_in
                    and int(n_min_right_counter) < int(len(right_vt_min))
                    and w_in[imin] < a_abs_min_in
                ):
                    right_vt_min[int(n_min_right_counter)] = imin
                    n_min_right_counter += 1
                    imax = i
                    find_max = True

    # set output
    # left search
    if search_direction == 0:
        n_max_out[0] = n_max_left_counter
        n_min_out[0] = n_min_left_counter
        vt_max_out[:] = left_vt_max
        vt_min_out[:] = left_vt_min

    # right search
    elif search_direction == 1:
        n_max_out[0] = n_max_right_counter
        n_min_out[0] = n_min_right_counter
        vt_max_out[:] = right_vt_max
        vt_min_out[:] = right_vt_min

    # conservative search (only extrema found in both directions)
    elif search_direction == 2:
        # sort the right search result. Left search should be already sorted
        right_vt_max = np.sort(right_vt_max)
        right_vt_min = np.sort(right_vt_min)

        rge_right = (right_vt_max[~np.isnan(right_vt_max)]).astype(np.int_)
        rge_left = (left_vt_max[~np.isnan(left_vt_max)]).astype(np.int_)

        # only continue if both arrays have something in them
        if len(rge_right) > 0 and len(rge_left) > 0:
            r_max = int(rge_right[-1])
            r_min = int(rge_right[0])
            rge = r_max - r_min

            # coincidence mask
            coin_mask = np.zeros(len(rge_left), dtype=np.bool_)

            # helper array: 1 if integer exists in rge_right
            helper_ar = np.zeros(rge + 1, dtype=np.bool_)
            helper_ar[rge_right - r_min] = 1

            # masking
            # only integers in [r_min,r_max] could be in coincidence
            basic_mask = (rge_left <= r_max) & (rge_left >= r_min)

            # take candidates of rge_left surviving the basic mask,
            # shift by rmin to get the indices for the integer mask.
            # Apply basic mask to coincidenc mask and indexed integer mask.
            coin_mask[basic_mask] = helper_ar[rge_left[basic_mask] - r_min]

            # Apply coincidence mask to rge_left & fill into output
            rge_left = rge_left[coin_mask]
            vt_max_out[: len(rge_left)] = rge_left
            n_max_out[0] = len(rge_left)
        else:
            n_max_out[0] = 0
        # Do the same for the minima
        rge_right = (right_vt_max[~np.isnan(right_vt_min)]).astype(np.int_)
        rge_left = (left_vt_max[~np.isnan(left_vt_min)]).astype(np.int_)

        # only continue if both arrays have something in them
        if len(rge_right) > 0 and len(rge_left) > 0:
            r_max = int(rge_right[-1])
            r_min = int(rge_right[0])
            rge = r_max - r_min

            # coincidence mask
            coin_mask = np.zeros(len(rge_left), dtype=np.bool_)

            # helper array: 1 if integer exists in rge_right
            helper_ar = np.zeros(rge + 1, dtype=np.bool_)
            helper_ar[rge_right - r_min] = 1

            # masking
            # only integers in [r_min,r_max] could be in coincidence
            basic_mask = (rge_left <= r_max) & (rge_left >= r_min)

            # take candidates of rge_left surviving the basic mask,
            # shift by rmin to get the indices for the integer mask.
            # Apply basic mask to coincidenc mask and indexed integer mask.
            coin_mask[basic_mask] = helper_ar[rge_left[basic_mask] - r_min]

            # Apply coincidence mask to rge_left & fill into output
            rge_left = rge_left[coin_mask]
            vt_min_out[: len(rge_left)] = rge_left
            n_min_out[0] = len(rge_left)
        else:
            n_min_out[0] = 0

    # aggressive search (extrema found in either directions)
    elif search_direction == 3:
        both = np.unique(np.append(left_vt_max, right_vt_max))
        if len(vt_max_out) <= len(both):
            vt_max_out[:] = both[: len(vt_max_out)]
            n_max_out[0] = len(vt_max_out[~np.isnan(vt_max_out)])
        else:
            vt_max_out[: len(both)] = both
            n_max_out[0] = len(both[~np.isnan(both)])

        both = np.unique(np.append(left_vt_min, right_vt_min))
        if len(vt_min_out) <= len(both):
            vt_min_out[:] = both[: len(vt_min_out)]
            n_min_out[0] = len(vt_min_out[~np.isnan(vt_min_out)])
        else:
            vt_min_out[: len(both)] = both
            n_min_out[0] = len(both[~np.isnan(both)])

    else:
        raise DSPFatal("search direction type not found.")



r"""
The pygama signal processing framework is contained in the :mod:`.dsp`
submodule. This code is responsible for running a variety of discrete signal
processors on data, producing tier 2 (dsp) files from tier 1 (raw) files. The
main contents of this submodule are:

* :mod:`.processors`: A collection of Numba functions that perform individual
  DSP transforms and reductions on our data. Here is where contributors will
  write new processors for inclusion in the analysis. Available processors
  include all :class:`numpy.ufunc`\ s as well.
* :class:`.ProcessingChain`: A class that manages and efficiently runs a list
  of DSP processors
* :func:`.build_processing_chain`: A function that builds a :class:`.ProcessingChain`
  using LH5-formatted input and output files, and a JSON configuration file
* :func:`.build_dsp`: A function that runs :func:`.build_processing_chain` to build a
  :class:`.ProcessingChain` from a JSON config file and then processes an input
  file and writes into an output file, using the LH5 file format
"""

from pygama.dsp.build_dsp import build_dsp
from pygama.dsp.processing_chain import ProcessingChain, build_processing_chain

__all__ = ["build_dsp", "ProcessingChain", "build_processing_chain"]



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32[:], float32[:])",
        "void(float64[:], float64, float64[:], float64[:])",
    ],
    "(n),()->(),()",
    **nb_kwargs,
)
def saturation(
    w_in: np.ndarray, bit_depth_in: int, n_lo_out: int, n_hi_out: int
) -> None:
    """Count the number of samples in the waveform that are saturated at the
    minimum and maximum possible values based on the bit depth.

    Parameters
    ----------
    w_in
        the input waveform
    bit_depth_in
        the bit depth of the analog-to-digital converter
    n_lo_out
        the output number of samples at the minimum
    n_hi_out
        the output number of samples at the maximum

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "sat_lo, sat_hi": {
            "function": "saturation",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "16", "sat_lo", "sat_hi"],
            "unit": "ADC"
        }
    """
    n_lo_out[0] = np.nan
    n_hi_out[0] = np.nan

    if np.isnan(w_in).any() or np.isnan(bit_depth_in):
        return

    if not np.floor(bit_depth_in) == bit_depth_in:
        raise DSPFatal("The bit depth is not an integer")

    if bit_depth_in <= 0:
        raise DSPFatal("The bit depth is not positive")

    n_lo_out[0] = 0
    n_hi_out[0] = 0
    for i in range(0, len(w_in), 1):
        if w_in[i] == 0:
            n_lo_out[0] += 1
        elif w_in[i] == np.power(2, int(bit_depth_in)) - int(bit_depth_in):
            n_hi_out[0] += 1



import numpy as np
import pytest

from pygama import lgdo
from pygama.raw.fc.fc_event_decoder import FCEventDecoder
from pygama.raw.raw_buffer import RawBuffer


@pytest.fixture(scope="module")
def event_rbkd(fcio_obj, fcio_config):
    decoder = FCEventDecoder()
    decoder.set_file_config(fcio_config)

    # get just one record (because size=1, see below) and check if it's a (sparse)event
    assert fcio_obj.get_record() == 3 or fcio_obj.get_record() == 6

    # build raw buffer for each channel in the FC trace list
    rbkd = {}
    for i in fcio_obj.tracelist:
        rbkd[i] = RawBuffer(lgdo=decoder.make_lgdo(size=1))

    # decode packet into the lgdo's and check if the buffer is full
    assert decoder.decode_packet(fcio=fcio_obj, evt_rbkd=rbkd, packet_id=69) is True

    # # check compression settings (here before any LH5Store.write_object() call
    # assert "compression" in rbkd[0].lgdo["packet_id"].attrs
    # assert "compression" in rbkd[0].lgdo["waveform"].values.attrs

    return rbkd


def test_decoding_and_compression_attrs(event_rbkd):
    assert event_rbkd != {}


def test_data_types(event_rbkd):
    for _, v in event_rbkd.items():
        tbl = v.lgdo
        assert isinstance(tbl, lgdo.Struct)
        assert isinstance(tbl["packet_id"], lgdo.Array)
        assert isinstance(tbl["eventnumber"], lgdo.Array)
        assert isinstance(tbl["timestamp"], lgdo.Array)
        assert isinstance(tbl["runtime"], lgdo.Array)
        assert isinstance(tbl["numtraces"], lgdo.Array)
        assert isinstance(tbl["tracelist"], lgdo.VectorOfVectors)
        assert isinstance(tbl["baseline"], lgdo.Array)
        assert isinstance(tbl["daqenergy"], lgdo.Array)
        assert isinstance(tbl["channel"], lgdo.Array)
        assert isinstance(tbl["ts_pps"], lgdo.Array)
        assert isinstance(tbl["ts_ticks"], lgdo.Array)
        assert isinstance(tbl["ts_maxticks"], lgdo.Array)
        assert isinstance(tbl["mu_offset_sec"], lgdo.Array)
        assert isinstance(tbl["mu_offset_usec"], lgdo.Array)
        assert isinstance(tbl["to_master_sec"], lgdo.Array)
        assert isinstance(tbl["delta_mu_usec"], lgdo.Array)
        assert isinstance(tbl["abs_delta_mu_usec"], lgdo.Array)
        assert isinstance(tbl["to_start_sec"], lgdo.Array)
        assert isinstance(tbl["to_start_usec"], lgdo.Array)
        assert isinstance(tbl["dr_start_pps"], lgdo.Array)
        assert isinstance(tbl["dr_start_ticks"], lgdo.Array)
        assert isinstance(tbl["dr_stop_pps"], lgdo.Array)
        assert isinstance(tbl["dr_stop_ticks"], lgdo.Array)
        assert isinstance(tbl["dr_maxticks"], lgdo.Array)
        assert isinstance(tbl["deadtime"], lgdo.Array)
        assert isinstance(tbl["waveform"], lgdo.Struct)
        assert isinstance(tbl["waveform"]["t0"], lgdo.Array)
        assert isinstance(tbl["waveform"]["dt"], lgdo.Array)
        assert isinstance(tbl["waveform"]["values"], lgdo.ArrayOfEqualSizedArrays)


def test_values(event_rbkd, fcio_obj):
    fc = fcio_obj
    for ch in fc.tracelist:
        loc = event_rbkd[ch].loc - 1
        tbl = event_rbkd[ch].lgdo

        assert event_rbkd[ch].fill_safety == fc.numtraces

        assert tbl["packet_id"].nda[loc] == 69
        assert tbl["eventnumber"].nda[loc] == fc.eventnumber
        assert tbl["timestamp"].nda[loc] == fc.eventtime
        assert tbl["runtime"].nda[loc] == fc.runtime
        assert tbl["numtraces"].nda[loc] == fc.numtraces

        # custom logic for VectorOfVectors
        start = 0 if loc == 0 else tbl["tracelist"].cumulative_length.nda[loc - 1]
        stop = start + len(fc.tracelist)
        assert np.array_equal(
            tbl["tracelist"].flattened_data.nda[start:stop], fc.tracelist
        )

        assert np.array_equal(tbl["baseline"].nda[loc], fc.baseline[ch])
        assert np.array_equal(tbl["daqenergy"].nda[loc], fc.daqenergy[ch])
        assert tbl["channel"].nda[loc] == ch
        assert tbl["ts_pps"].nda[loc] == fc.timestamp_pps
        assert tbl["ts_ticks"].nda[loc] == fc.timestamp_ticks
        assert tbl["ts_maxticks"].nda[loc] == fc.timestamp_maxticks
        assert tbl["mu_offset_sec"].nda[loc] == fc.timeoffset_mu_sec
        assert tbl["mu_offset_usec"].nda[loc] == fc.timeoffset_mu_usec
        assert tbl["to_master_sec"].nda[loc] == fc.timeoffset_master_sec
        assert tbl["delta_mu_usec"].nda[loc] == fc.timeoffset_dt_mu_usec
        assert tbl["abs_delta_mu_usec"].nda[loc] == fc.timeoffset_abs_mu_usec
        assert tbl["to_start_sec"].nda[loc] == fc.timeoffset_start_sec
        assert tbl["to_start_usec"].nda[loc] == fc.timeoffset_start_usec
        assert tbl["dr_start_pps"].nda[loc] == fc.deadregion_start_pps
        assert tbl["dr_start_ticks"].nda[loc] == fc.deadregion_start_ticks
        assert tbl["dr_stop_pps"].nda[loc] == fc.deadregion_stop_pps
        assert tbl["dr_stop_ticks"].nda[loc] == fc.deadregion_stop_ticks
        assert tbl["dr_maxticks"].nda[loc] == fc.deadregion_maxticks
        assert tbl["deadtime"].nda[loc] == fc.deadtime
        assert tbl["waveform"]["t0"].nda[loc] == 0
        assert tbl["waveform"]["dt"].nda[loc] == 16
        assert np.array_equal(tbl["waveform"]["values"].nda[loc], fc.traces[ch])



from __future__ import annotations

import logging

import numpy as np
import xmltodict

from pygama import lgdo

log = logging.getLogger(__name__)


max_number_channels = 8
max_number_of_boards = 2  # overload this value if you are using more boards and don't want to supply a config file!
DT5730_sample_freq = 0.5  # sample/ns, because wf_len is stored in ns in the xml file
DT5725_sample_freq = 0.25  # sample/ns


def compass_config_to_struct(
    compass_config_file: str = None, wf_len: int = None
) -> lgdo.Struct:
    """
    Read run-level data from a CoMPASS configuration file.

    Parameters
    ----------
    compass_config_file
        An xml config file for a CoMPASS data run

    Returns
    -------
    config_struct
        A struct of configuration data to pass to the streamer

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        {
            "boards" : {
                "0" : {
                    "modelName" : "DT5730",
                    "adcBitCount" : 14,
                    "sample_rate" : 0.5,
                    "wf_len" : 1000,
                    "channels": {
                        "0" : {
                            "Vpp": 2,
                        }
                    }
                }
            }
        }

    Notes
    -----
    First, if the CoMPASS xml config is not none, it is turned into a dictionary.
    If the CoMPASS config file is none, then a generic struct with the
    maximum number of channels and boards is returned, in order to create the
    most generic raw buffer to fill.
    """
    # If the config file is not none, read in the xml and convert it to dict
    if compass_config_file is not None:
        with open(compass_config_file) as xml_file:
            data_dict = xmltodict.parse(xml_file.read())

        # Initialize a struct to hold the config data
        config_struct = lgdo.Struct()
        config_struct.add_field("boards", lgdo.Struct())

        # get the number of boards enabled
        # "configuration/board" is a dictionary if there is only one board
        if isinstance(data_dict["configuration"]["board"], dict):
            config_struct["boards"].add_field("0", lgdo.Struct())
            # Read in board and adc
            config_struct["boards"]["0"].add_field(
                "modelName",
                lgdo.Scalar(data_dict["configuration"]["board"]["modelName"]),
            )
            config_struct["boards"]["0"].add_field(
                "adcBitCount",
                lgdo.Scalar(data_dict["configuration"]["board"]["adcBitCount"]),
            )

            # Write the sampling rate, in samples/ns and store for conversion of the wf_len
            if config_struct["boards"]["0"]["modelName"].value == "DT5730":
                config_struct["boards"]["0"].add_field(
                    "sample_rate", lgdo.Scalar(DT5730_sample_freq)
                )
            elif config_struct["boards"]["0"]["modelName"].value == "DT5725":
                config_struct["boards"]["0"].add_field(
                    "sample_rate", lgdo.Scalar(DT5725_sample_freq)
                )
            else:
                raise NotImplementedError(
                    f'warning! no sample rate defined yet for board {config_struct["boards"][0]["modelName"].value}'
                )
                config_struct["boards"]["0"].add_field("sample_rate", lgdo.Scalar(1))

            # The parameters tag is a list of dictionaries, turn it into one dictionary
            param_dict = {}
            for item in data_dict["configuration"]["board"]["parameters"]["entry"]:
                param_name = item["key"]
                param_dict[param_name] = item["value"]

            # Read in wf_len
            if "SRV_PARAM_RECLEN" in param_dict.keys():
                # all xml values are stored with the key "#text" in the config_file
                # the wf_len are stored in ns, but we need the sample length to decode the packets, so convert using the sample rate
                config_struct["boards"]["0"].add_field(
                    "wf_len",
                    lgdo.Scalar(
                        float(param_dict["SRV_PARAM_RECLEN"]["value"]["#text"])
                        * float(config_struct["boards"]["0"]["sample_rate"].value)
                    ),
                )
            else:
                raise RuntimeError("wf_len not in config file!")

            # Now, loop through the channels and update the struct
            config_struct["boards"]["0"].add_field("channels", lgdo.Struct())

            for channel_dictionary in data_dict["configuration"]["board"]["channel"]:
                channel_number = channel_dictionary["index"]
                # first check to make sure that the channel wasn't disabled from the compass interface,
                # the xml file would then have <index>channel_no</index></values>
                if channel_dictionary["values"] is None:
                    continue
                # if channel_dictionary["values"]["entry"] is a dictionary, then that channel is not enabled
                if isinstance(channel_dictionary["values"]["entry"], dict):
                    continue
                # initialize the sub struct
                else:
                    config_struct["boards"]["0"]["channels"].add_field(
                        str(channel_number), lgdo.Struct()
                    )

                for x in channel_dictionary["values"]["entry"]:
                    # if a value isn't present, then the key is just set to the default in the param_dict
                    if (x["key"] in param_dict.keys()) and ("value" not in x.keys()):
                        config_struct["boards"]["0"]["channels"][
                            str(channel_number)
                        ].add_field(
                            str(x["key"]),
                            lgdo.Scalar(param_dict[x["key"]]["value"]["#text"]),
                        )

                    else:
                        config_struct["boards"]["0"]["channels"][
                            str(channel_number)
                        ].add_field(str(x["key"]), lgdo.Scalar(x["value"]["#text"]))

        # if there is more than one board, loop over each board and get the values
        if isinstance(data_dict["configuration"]["board"], list):
            for i in range(len(data_dict["configuration"]["board"])):
                config_struct["boards"].add_field(str(i), lgdo.Struct())
                # Read in board and adc
                config_struct["boards"][str(i)].add_field(
                    "modelName",
                    lgdo.Scalar(data_dict["configuration"]["board"][i]["modelName"]),
                )

                config_struct["boards"][str(i)].add_field(
                    "adcBitCount",
                    lgdo.Scalar(data_dict["configuration"]["board"][i]["adcBitCount"]),
                )

                # Write the sampling rate, in samples/ns and store for conversion of the wf_len
                if config_struct["boards"][str(i)]["modelName"].value == "DT5730":
                    config_struct["boards"][str(i)].add_field(
                        "sample_rate", lgdo.Scalar(DT5730_sample_freq)
                    )
                elif config_struct["boards"][str(i)]["modelName"].value == "DT5725":
                    config_struct["boards"][str(i)].add_field(
                        "sample_rate", lgdo.Scalar(DT5725_sample_freq)
                    )
                else:
                    raise NotImplementedError(
                        f'warning! no sample rate defined yet for board {config_struct["boards"][i]["modelName"].value}'
                    )
                    config_struct["boards"][str(i)].add_field(
                        "sample_rate", lgdo.Scalar(1)
                    )

                # The parameters tag is a list of dictionaries, turn it into one dictionary
                param_dict = {}
                for item in data_dict["configuration"]["board"][i]["parameters"][
                    "entry"
                ]:
                    param_name = item["key"]
                    param_dict[param_name] = item["value"]

                # Read in wf_len
                if "SRV_PARAM_RECLEN" in param_dict.keys():
                    # all xml values are stored with the key "#text" in the config_file
                    # the wf_len are stored in ns, but we need the sample length to decode the packets, so convert using the sample rate
                    config_struct["boards"][str(i)].add_field(
                        "wf_len",
                        lgdo.Scalar(
                            float(param_dict["SRV_PARAM_RECLEN"]["value"]["#text"])
                            * float(
                                config_struct["boards"][str(i)]["sample_rate"].value
                            )
                        ),
                    )  # all xml values are stored with the key "#text" in the config_file
                else:
                    raise RuntimeError("wf_len not in config file!")

                config_struct["boards"][str(i)].add_field("channels", lgdo.Struct())

                # Now, loop through the channels and update the dictionary
                for channel_dictionary in data_dict["configuration"]["board"][i][
                    "channel"
                ]:
                    channel_number = channel_dictionary["index"]
                    # if channel_dictionary["values"]["entry"] is a dictionary, then that channel is not enabled
                    if isinstance(channel_dictionary["values"]["entry"], dict):
                        continue
                    # initialize the sub struct
                    else:
                        config_struct["boards"][str(i)]["channels"].add_field(
                            str(channel_number), lgdo.Struct()
                        )

                    for x in channel_dictionary["values"]["entry"]:
                        # if a value isn't present, then the key is just set to the default in the param_dict
                        if (x["key"] in param_dict.keys()) and (
                            "value" not in x.keys()
                        ):
                            config_struct["boards"][str(i)]["channels"][
                                str(channel_number)
                            ].add_field(
                                str(x["key"]),
                                lgdo.Scalar(param_dict[x["key"]]["value"]["#text"]),
                            )
                        else:
                            config_struct["boards"][str(i)]["channels"][
                                str(channel_number)
                            ].add_field(str(x["key"]), lgdo.Scalar(x["value"]["#text"]))

        return config_struct

    # if the config file is none, return a generic struct with the most channels and boards enabled possible, in order to create a large enough raw buffer
    else:
        config_struct = lgdo.Struct()
        config_struct.add_field("boards", lgdo.Struct())
        for i in range(max_number_of_boards):
            config_struct["boards"].add_field(str(i), lgdo.Struct())

            # Set board and adc to none
            config_struct["boards"][str(i)].add_field("modelName", lgdo.Scalar(np.nan))
            config_struct["boards"][str(i)].add_field(
                "adcBitCount", lgdo.Scalar(np.nan)
            )
            config_struct["boards"][str(i)].add_field(
                "sample_rate", lgdo.Scalar(np.nan)
            )

            # set the wf_len
            if wf_len is not None:
                config_struct["boards"][str(i)].add_field(
                    "wf_len", lgdo.Scalar(float(wf_len))
                )
            else:
                raise RuntimeError(
                    "wf_len not passed correctly when config file is absent!"
                )
            config_struct["boards"][str(i)].add_field("channels", lgdo.Struct())
            for j in range(max_number_channels):
                # initialize the sub-dict
                config_struct["boards"][str(i)]["channels"].add_field(
                    str(j), lgdo.Scalar(np.nan)
                )

        return config_struct



import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs

from .fixed_time_pickoff import fixed_time_pickoff


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:])",
    ],
    "(n),(m),(m)",
    **nb_kwargs,
    forceobj=True
)
def multi_a_filter(w_in, vt_maxs_in, va_max_out):
    """Finds the maximums in a waveform and returns the amplitude of the wave
    at those points.

    Parameters
    ----------
    w_in
        the array of data within which amplitudes of extrema will be found.
    vt_maxs_in
        the array of max positions for each waveform.
    va_max_out
        an array (in-place filled) of the amplitudes of the maximums of the waveform.
    """

    # Initialize output parameters

    va_max_out[:] = np.nan

    # Check inputs

    if np.isnan(w_in).any():
        return

    if not len(vt_maxs_in) < len(w_in):
        raise DSPFatal(
            "The length of your return array must be smaller than the length of your waveform"
        )

    fixed_time_pickoff(w_in, vt_maxs_in, ord("i"), va_max_out)



import logging
import os

import h5py
import numpy as np
import pandas as pd
import pytest

import pygama.lgdo as lgdo
import pygama.lgdo.lh5_store as lh5
from pygama.lgdo import compression
from pygama.lgdo.compression import RadwareSigcompress
from pygama.lgdo.lh5_store import LH5Store


@pytest.fixture(scope="module")
def lgnd_file(lgnd_test_data):
    return lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5")


def test_init():
    LH5Store()


def test_gimme_file(lgnd_file):
    store = LH5Store(keep_open=True)

    f = store.gimme_file(lgnd_file)
    assert isinstance(f, h5py.File)
    assert store.files[lgnd_file] == f

    with pytest.raises(FileNotFoundError):
        store.gimme_file("non-existent-file")


def test_gimme_group(lgnd_file):
    f = h5py.File(lgnd_file)
    store = LH5Store()
    g = store.gimme_group("/geds", f)
    assert isinstance(g, h5py.Group)

    f = h5py.File("/tmp/testfile.lh5", mode="w")
    g = store.gimme_group("/geds", f, grp_attrs={"attr1": 1}, overwrite=True)
    assert isinstance(g, h5py.Group)


def test_show(lgnd_file):
    lh5.show(lgnd_file)
    lh5.show(lgnd_file, "/geds/raw")
    lh5.show(lgnd_file, "geds/raw")


def test_ls(lgnd_file):
    assert lh5.ls(lgnd_file) == ["geds"]
    assert lh5.ls(lgnd_file, "/*/raw") == ["geds/raw"]
    assert lh5.ls(lgnd_file, "geds/raw/") == [
        "geds/raw/baseline",
        "geds/raw/channel",
        "geds/raw/energy",
        "geds/raw/ievt",
        "geds/raw/numtraces",
        "geds/raw/packet_id",
        "geds/raw/timestamp",
        "geds/raw/tracelist",
        "geds/raw/waveform",
        "geds/raw/wf_max",
        "geds/raw/wf_std",
    ]


def test_load_nda(lgnd_file):
    nda = lh5.load_nda(
        [lgnd_file, lgnd_file],
        ["baseline", "waveform/values"],
        lh5_group="/geds/raw",
        idx_list=[[1, 3, 5], [2, 6, 7]],
    )

    assert isinstance(nda, dict)
    assert isinstance(nda["baseline"], np.ndarray)
    assert nda["baseline"].shape == (6,)
    assert isinstance(nda["waveform/values"], np.ndarray)
    assert nda["waveform/values"].shape == (6, 5592)


def test_load_dfs(lgnd_file):
    dfs = lh5.load_dfs(
        [lgnd_file, lgnd_file],
        ["baseline", "waveform/t0"],
        lh5_group="/geds/raw",
        idx_list=[[1, 3, 5], [2, 6, 7]],
    )

    assert isinstance(dfs, pd.DataFrame)


@pytest.fixture(scope="module")
def lh5_file():
    store = LH5Store()

    struct = lgdo.Struct()
    struct.add_field("scalar", lgdo.Scalar(value=10, attrs={"sth": 1}))
    struct.add_field("array", lgdo.Array(nda=np.array([1, 2, 3, 4, 5])))
    struct.add_field(
        "aoesa",
        lgdo.ArrayOfEqualSizedArrays(shape=(5, 5), dtype=np.float32, fill_val=42),
    )
    struct.add_field(
        "vov",
        lgdo.VectorOfVectors(
            flattened_data=lgdo.Array(
                nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])
            ),
            cumulative_length=lgdo.Array(nda=np.array([2, 5, 6, 10, 13])),
            attrs={"myattr": 2},
        ),
    )

    struct.add_field(
        "voev",
        lgdo.VectorOfEncodedVectors(
            encoded_data=lgdo.VectorOfVectors(
                flattened_data=lgdo.Array(
                    nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])
                ),
                cumulative_length=lgdo.Array(nda=np.array([2, 5, 6, 10, 13])),
            ),
            decoded_size=lgdo.Array(shape=5, fill_val=6),
        ),
    )

    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4]), attrs={"attr": 9}),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8]), attrs={"compression": "gzip"}),
    }

    struct.add_field("table", lgdo.Table(col_dict=col_dict, attrs={"stuff": 5}))

    struct.add_field(
        "wftable",
        lgdo.WaveformTable(
            t0=lgdo.Array(np.zeros(10)),
            dt=lgdo.Array(np.full(10, fill_value=1)),
            values=lgdo.ArrayOfEqualSizedArrays(
                shape=(10, 1000), dtype=np.uint16, fill_val=100, attrs={"custom": 8}
            ),
        ),
    )

    struct.add_field(
        "wftable_enc",
        lgdo.WaveformTable(
            t0=lgdo.Array(np.zeros(10)),
            dt=lgdo.Array(np.full(10, fill_value=1)),
            values=compression.encode(
                struct["wftable"].values, codec=RadwareSigcompress(codec_shift=-32768)
            ),
        ),
    )

    store.write_object(
        struct,
        "struct",
        "/tmp/tmp-pygama-lgdo-types.lh5",
        group="/data",
        start_row=1,
        n_rows=3,
        wo_mode="overwrite_file",
    )

    store.write_object(
        struct,
        "struct_full",
        "/tmp/tmp-pygama-lgdo-types.lh5",
        group="/data",
        wo_mode="append",
    )

    return "/tmp/tmp-pygama-lgdo-types.lh5"


def test_write_objects(lh5_file):
    pass


def test_read_n_rows(lh5_file):
    store = LH5Store()
    assert store.read_n_rows("/data/struct_full/aoesa", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/array", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/scalar", lh5_file) is None
    assert store.read_n_rows("/data/struct_full/table", lh5_file) == 4
    assert store.read_n_rows("/data/struct_full/voev", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/vov", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/wftable", lh5_file) == 10
    assert store.read_n_rows("/data/struct_full/wftable_enc/values", lh5_file) == 10


def test_get_buffer(lh5_file):
    store = LH5Store()
    buf = store.get_buffer("/data/struct_full/wftable_enc", lh5_file)
    assert isinstance(buf.values, lgdo.ArrayOfEqualSizedArrays)


def test_read_scalar(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/scalar", lh5_file)
    assert isinstance(lh5_obj, lgdo.Scalar)
    assert lh5_obj.value == 10
    assert n_rows == 1
    assert lh5_obj.attrs["sth"] == 1


def test_read_array(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/array", lh5_file)
    assert isinstance(lh5_obj, lgdo.Array)
    assert (lh5_obj.nda == np.array([2, 3, 4])).all()
    assert n_rows == 3


def test_read_array_fancy_idx(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object(
        "/data/struct_full/array", lh5_file, idx=[0, 3, 4]
    )
    assert isinstance(lh5_obj, lgdo.Array)
    assert (lh5_obj.nda == np.array([1, 4, 5])).all()
    assert n_rows == 3


def test_read_vov(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/vov", lh5_file)
    assert isinstance(lh5_obj, lgdo.VectorOfVectors)

    desired = [np.array([3, 4, 5]), np.array([2]), np.array([4, 8, 9, 7])]

    for i in range(len(desired)):
        assert (desired[i] == list(lh5_obj)[i]).all()

    assert n_rows == 3
    assert lh5_obj.attrs["myattr"] == 2


def test_read_vov_fancy_idx(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct_full/vov", lh5_file, idx=[0, 2])
    assert isinstance(lh5_obj, lgdo.VectorOfVectors)

    desired = [np.array([1, 2]), np.array([2])]

    for i in range(len(desired)):
        assert (desired[i] == list(lh5_obj)[i]).all()

    assert n_rows == 2


def test_read_voev(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/voev", lh5_file, decompress=False)
    assert isinstance(lh5_obj, lgdo.VectorOfEncodedVectors)

    desired = [np.array([3, 4, 5]), np.array([2]), np.array([4, 8, 9, 7])]

    for i in range(len(desired)):
        assert (desired[i] == lh5_obj[i][0]).all()

    assert n_rows == 3

    lh5_obj, n_rows = store.read_object(
        "/data/struct/voev", [lh5_file, lh5_file], decompress=False
    )
    assert isinstance(lh5_obj, lgdo.VectorOfEncodedVectors)
    assert n_rows == 6


def test_read_voev_fancy_idx(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object(
        "/data/struct_full/voev", lh5_file, idx=[0, 2], decompress=False
    )
    assert isinstance(lh5_obj, lgdo.VectorOfEncodedVectors)

    desired = [np.array([1, 2]), np.array([2])]

    for i in range(len(desired)):
        assert (desired[i] == lh5_obj[i][0]).all()

    assert n_rows == 2


def test_read_aoesa(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/aoesa", lh5_file)
    assert isinstance(lh5_obj, lgdo.ArrayOfEqualSizedArrays)
    assert (lh5_obj.nda == np.full((3, 5), fill_value=42)).all()


def test_read_table(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/table", lh5_file)
    assert isinstance(lh5_obj, lgdo.Table)
    assert n_rows == 3

    lh5_obj, n_rows = store.read_object("/data/struct/table", [lh5_file, lh5_file])
    assert n_rows == 6
    assert lh5_obj.attrs["stuff"] == 5
    assert lh5_obj["a"].attrs["attr"] == 9


def test_read_hdf5_compressed_data(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/table", lh5_file)

    assert "compression" not in lh5_obj["b"].attrs
    with h5py.File(lh5_file) as h5f:
        assert h5f["/data/struct/table/b"].compression == "gzip"
        assert h5f["/data/struct/table/a"].compression is None


def test_read_wftable(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/data/struct/wftable", lh5_file)
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert n_rows == 3

    lh5_obj, n_rows = store.read_object("/data/struct/wftable", [lh5_file, lh5_file])
    assert n_rows == 6
    assert lh5_obj.values.attrs["custom"] == 8


def test_read_wftable_encoded(lh5_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object(
        "/data/struct/wftable_enc", lh5_file, decompress=False
    )
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert isinstance(lh5_obj.values, lgdo.ArrayOfEncodedEqualSizedArrays)
    assert n_rows == 3
    assert lh5_obj.values.attrs["codec"] == "radware_sigcompress"
    assert "codec_shift" in lh5_obj.values.attrs

    lh5_obj, n_rows = store.read_object("/data/struct/wftable_enc", lh5_file)
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert isinstance(lh5_obj.values, lgdo.ArrayOfEqualSizedArrays)
    assert n_rows == 3

    lh5_obj, n_rows = store.read_object("/data/struct/wftable_enc/values", lh5_file)
    assert isinstance(lh5_obj, lgdo.ArrayOfEqualSizedArrays)
    assert n_rows == 3

    lh5_obj, n_rows = store.read_object(
        "/data/struct/wftable_enc", [lh5_file, lh5_file], decompress=False
    )
    assert n_rows == 6

    lh5_obj, n_rows = store.read_object(
        "/data/struct/wftable_enc", [lh5_file, lh5_file], decompress=True
    )
    assert n_rows == 6


def test_read_with_field_mask(lh5_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object(
        "/data/struct_full", lh5_file, field_mask=["array"]
    )
    assert list(lh5_obj.keys()) == ["array"]

    lh5_obj, n_rows = store.read_object(
        "/data/struct_full", lh5_file, field_mask=("array", "table")
    )
    assert list(lh5_obj.keys()) == ["array", "table"]

    lh5_obj, n_rows = store.read_object(
        "/data/struct_full", lh5_file, field_mask={"array": True}
    )
    assert list(lh5_obj.keys()) == ["array"]

    lh5_obj, n_rows = store.read_object(
        "/data/struct_full", lh5_file, field_mask={"vov": False, "voev": False}
    )
    assert list(lh5_obj.keys()) == [
        "scalar",
        "array",
        "aoesa",
        "table",
        "wftable",
        "wftable_enc",
    ]


def test_read_lgnd_array(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object("/geds/raw/baseline", lgnd_file)
    assert isinstance(lh5_obj, lgdo.Array)
    assert n_rows == 100
    assert len(lh5_obj) == 100

    lh5_obj, n_rows = store.read_object("/geds/raw/waveform/values", lgnd_file)
    assert isinstance(lh5_obj, lgdo.ArrayOfEqualSizedArrays)


def test_read_lgnd_array_fancy_idx(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object(
        "/geds/raw/baseline", lgnd_file, idx=[2, 4, 6, 9, 11, 16, 68]
    )
    assert isinstance(lh5_obj, lgdo.Array)
    assert n_rows == 7
    assert len(lh5_obj) == 7
    assert (lh5_obj.nda == [13508, 14353, 14525, 14341, 15079, 11675, 13995]).all()


def test_read_lgnd_vov(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object("/geds/raw/tracelist", lgnd_file)
    assert isinstance(lh5_obj, lgdo.VectorOfVectors)
    assert n_rows == 100
    assert len(lh5_obj) == 100


def test_read_lgnd_vov_fancy_idx(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object(
        "/geds/raw/tracelist", lgnd_file, idx=[2, 4, 6, 9, 11, 16, 68]
    )
    assert isinstance(lh5_obj, lgdo.VectorOfVectors)
    assert n_rows == 7
    assert len(lh5_obj) == 7
    assert (lh5_obj.cumulative_length.nda == [1, 2, 3, 4, 5, 6, 7]).all()
    assert (lh5_obj.flattened_data.nda == [40, 60, 64, 60, 64, 28, 60]).all()


def test_read_array_concatenation(lgnd_file):
    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/geds/raw/baseline", [lgnd_file, lgnd_file])
    assert isinstance(lh5_obj, lgdo.Array)
    assert n_rows == 200
    assert len(lh5_obj) == 200


def test_read_lgnd_waveform_table(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object("/geds/raw/waveform", lgnd_file)
    assert isinstance(lh5_obj, lgdo.WaveformTable)

    lh5_obj, n_rows = store.read_object(
        "/geds/raw/waveform",
        lgnd_file,
        start_row=10,
        n_rows=10,
        field_mask=["t0", "dt"],
    )

    assert isinstance(lh5_obj, lgdo.Table)
    assert list(lh5_obj.keys()) == ["t0", "dt"]
    assert len(lh5_obj) == 10


def test_read_lgnd_waveform_table_fancy_idx(lgnd_file):
    store = LH5Store()

    lh5_obj, n_rows = store.read_object(
        "/geds/raw/waveform",
        lgnd_file,
        idx=[7, 9, 25, 27, 33, 38, 46, 52, 57, 59, 67, 71, 72, 82, 90, 92, 93, 94, 97],
    )
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert len(lh5_obj) == 19


@pytest.fixture(scope="module")
def enc_lgnd_file(lgnd_file):
    store = LH5Store()
    wft, n_rows = store.read_object("/geds/raw/waveform", lgnd_file)
    wft.values.attrs["compression"] = RadwareSigcompress(codec_shift=-32768)
    store.write_object(
        wft,
        "/geds/raw/waveform",
        "/tmp/tmp-pygama-compressed-wfs.lh5",
        wo_mode="overwrite_file",
    )
    return "/tmp/tmp-pygama-compressed-wfs.lh5"


def test_write_compressed_lgnd_waveform_table(enc_lgnd_file):
    pass


def test_read_compressed_lgnd_waveform_table(lgnd_file, enc_lgnd_file):
    store = LH5Store()
    wft, _ = store.read_object("/geds/raw/waveform", enc_lgnd_file)
    assert isinstance(wft.values, lgdo.ArrayOfEqualSizedArrays)
    assert "compression" not in wft.values.attrs


# First test that we can overwrite a table with the same name without deleting the original field
def test_write_object_overwrite_table_no_deletion(caplog):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    if os.path.exists("/tmp/write_object_overwrite_test.lh5"):
        os.remove("/tmp/write_object_overwrite_test.lh5")

    tb1 = lh5.Table(col_dict={"dset1": lh5.Array(np.zeros(10))})
    tb2 = lh5.Table(
        col_dict={"dset1": lh5.Array(np.ones(10))}
    )  # Same field name, different values
    store = LH5Store()
    store.write_object(tb1, "my_group", "/tmp/write_object_overwrite_test.lh5")
    store.write_object(
        tb2, "my_group", "/tmp/write_object_overwrite_test.lh5", wo_mode="overwrite"
    )  # Now, try to overwrite the same field

    # If the old field is deleted from the file before writing the new field, then we would get an extra debug statement
    assert "dset1 is not present in new table, deleting field" not in [
        rec.message for rec in caplog.records
    ]

    # Now, check that the data were overwritten
    tb_dat, _ = store.read_object("my_group", "/tmp/write_object_overwrite_test.lh5")
    assert np.array_equal(tb_dat["dset1"].nda, np.ones(10))


# Second: test that when we overwrite a table with fields with a different name, we delete the original field
def test_write_object_overwrite_table_with_deletion(caplog):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    if os.path.exists("/tmp/write_object_overwrite_test.lh5"):
        os.remove("/tmp/write_object_overwrite_test.lh5")

    tb1 = lh5.Table(col_dict={"dset1": lh5.Array(np.zeros(10))})
    tb2 = lh5.Table(
        col_dict={"dset2": lh5.Array(np.ones(10))}
    )  # Same field name, different values
    store = LH5Store()
    store.write_object(tb1, "my_group", "/tmp/write_object_overwrite_test.lh5")
    store.write_object(
        tb2, "my_group", "/tmp/write_object_overwrite_test.lh5", wo_mode="overwrite"
    )  # Now, try to overwrite with a different field

    # Now, check that the data were overwritten
    tb_dat, _ = store.read_object("my_group", "/tmp/write_object_overwrite_test.lh5")
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))

    # Also make sure that the first table's fields aren't lurking around the lh5 file!
    with h5py.File("/tmp/write_object_overwrite_test.lh5", "r") as lh5file:
        assert "dset1" not in list(lh5file["my_group"].keys())

    # Make sure the same behavior happens when we nest the table in a group
    if os.path.exists("/tmp/write_object_overwrite_test.lh5"):
        os.remove("/tmp/write_object_overwrite_test.lh5")

    tb1 = lh5.Table(col_dict={"dset1": lh5.Array(np.zeros(10))})
    tb2 = lh5.Table(
        col_dict={"dset2": lh5.Array(np.ones(10))}
    )  # Same field name, different values
    store = LH5Store()
    store.write_object(
        tb1, "my_table", "/tmp/write_object_overwrite_test.lh5", group="my_group"
    )
    store.write_object(
        tb2,
        "my_table",
        "/tmp/write_object_overwrite_test.lh5",
        group="my_group",
        wo_mode="overwrite",
    )  # Now, try to overwrite with a different field

    # Now, check that the data were overwritten
    tb_dat, _ = store.read_object(
        "my_group/my_table", "/tmp/write_object_overwrite_test.lh5"
    )
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))

    # Also make sure that the first table's fields aren't lurking around the lh5 file!
    with h5py.File("/tmp/write_object_overwrite_test.lh5", "r") as lh5file:
        assert "dset1" not in list(lh5file["my_group/my_table"].keys())


# Third: test that when we overwrite other LGDO classes
def test_write_object_overwrite_lgdo(caplog):
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    # Start with an lgdo.WaveformTable
    if os.path.exists("/tmp/write_object_overwrite_test.lh5"):
        os.remove("/tmp/write_object_overwrite_test.lh5")

    tb1 = lh5.WaveformTable(
        t0=np.zeros(10),
        t0_units="ns",
        dt=np.zeros(10),
        dt_units="ns",
        values=np.zeros((10, 10)),
        values_units="ADC",
    )
    tb2 = lh5.WaveformTable(
        t0=np.ones(10),
        t0_units="ns",
        dt=np.ones(10),
        dt_units="ns",
        values=np.ones((10, 10)),
        values_units="ADC",
    )  # Same field name, different values
    store = LH5Store()
    store.write_object(
        tb1, "my_table", "/tmp/write_object_overwrite_test.lh5", group="my_group"
    )
    store.write_object(
        tb2,
        "my_table",
        "/tmp/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        group="my_group",
    )

    # If the old field is deleted from the file before writing the new field, then we would get a debug statement
    assert "my_table is not present in new table, deleting field" not in [
        rec.message for rec in caplog.records
    ]

    # Now, check that the data were overwritten
    tb_dat, _ = store.read_object(
        "my_group/my_table", "/tmp/write_object_overwrite_test.lh5"
    )
    assert np.array_equal(tb_dat["values"].nda, np.ones((10, 10)))

    # Now try overwriting an array, and test the write_start argument
    array1 = lh5.Array(nda=np.zeros(10))
    array2 = lh5.Array(nda=np.ones(20))
    store = LH5Store()
    store.write_object(array1, "my_array", "/tmp/write_object_overwrite_test.lh5")
    store.write_object(
        array2,
        "my_array",
        "/tmp/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        write_start=5,
    )

    # Now, check that the data were overwritten
    array_dat, _ = store.read_object("my_array", "/tmp/write_object_overwrite_test.lh5")
    expected_out_array = np.append(np.zeros(5), np.ones(20))

    assert np.array_equal(array_dat.nda, expected_out_array)

    # Now try overwriting a scalar
    scalar1 = lh5.Scalar(0)
    scalar2 = lh5.Scalar(1)
    store = LH5Store()
    store.write_object(scalar1, "my_scalar", "/tmp/write_object_overwrite_test.lh5")
    store.write_object(
        scalar2,
        "my_scalar",
        "/tmp/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
    )

    # Now, check that the data were overwritten
    scalar_dat, _ = store.read_object(
        "my_scalar", "/tmp/write_object_overwrite_test.lh5"
    )

    assert scalar_dat.value == 1

    # Finally, try overwriting a vector of vectors
    vov1 = lh5.VectorOfVectors(listoflists=[np.zeros(1), np.ones(2), np.zeros(3)])
    vov2 = lh5.VectorOfVectors(listoflists=[np.ones(1), np.zeros(2), np.ones(3)])
    store = LH5Store()
    store.write_object(vov1, "my_vector", "/tmp/write_object_overwrite_test.lh5")
    store.write_object(
        vov2,
        "my_vector",
        "/tmp/write_object_overwrite_test.lh5",
        wo_mode="overwrite",
        write_start=1,
    )  # start overwriting the second list of lists

    vector_dat, _ = store.read_object(
        "my_vector", "/tmp/write_object_overwrite_test.lh5"
    )

    assert np.array_equal(vector_dat.cumulative_length.nda, [1, 2, 4, 7])
    assert np.array_equal(vector_dat.flattened_data.nda, [0, 1, 0, 0, 1, 1, 1])


# Test that when we try to overwrite an existing column in a table we fail
def test_write_object_append_column():
    # Try to append an array to a table
    if os.path.exists("/tmp/write_object_append_column_test.lh5"):
        os.remove("/tmp/write_object_append_column_test.lh5")

    array1 = lh5.Array(np.zeros(10))
    tb1 = lh5.Table(col_dict={"dset1`": lh5.Array(np.ones(10))})
    store = LH5Store()
    store.write_object(array1, "my_table", "/tmp/write_object_append_column_test.lh5")
    with pytest.raises(RuntimeError) as exc_info:
        store.write_object(
            tb1,
            "my_table",
            "/tmp/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column to an array

    assert exc_info.type is RuntimeError
    assert (
        exc_info.value.args[0] == "Trying to append columns to an object of type array"
    )

    # Try to append a table that has a same key as the old table
    if os.path.exists("/tmp/write_object_append_column_test.lh5"):
        os.remove("/tmp/write_object_append_column_test.lh5")

    tb1 = lh5.Table(
        col_dict={"dset1": lh5.Array(np.zeros(10)), "dset2": lh5.Array(np.zeros(10))}
    )
    tb2 = lh5.Table(
        col_dict={"dset2": lh5.Array(np.ones(10))}
    )  # Same field name, different values
    store = LH5Store()
    store.write_object(tb1, "my_table", "/tmp/write_object_append_column_test.lh5")
    with pytest.raises(ValueError) as exc_info:
        store.write_object(
            tb2,
            "my_table",
            "/tmp/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a same field

    assert exc_info.type is ValueError
    assert (
        exc_info.value.args[0]
        == "Can't append ['dset2'] column(s) to a table with the same field(s)"
    )

    # try appending a column that is larger than one that exists
    if os.path.exists("/tmp/write_object_append_column_test.lh5"):
        os.remove("/tmp/write_object_append_column_test.lh5")

    tb1 = lh5.Table(col_dict={"dset1": lh5.Array(np.zeros(10))})
    tb2 = lh5.Table(
        col_dict={"dset2": lh5.Array(np.ones(20))}
    )  # different field name, different size
    store = LH5Store()
    store.write_object(tb1, "my_table", "/tmp/write_object_append_column_test.lh5")
    with pytest.raises(ValueError) as exc_info:
        store.write_object(
            tb2,
            "my_table",
            "/tmp/write_object_append_column_test.lh5",
            wo_mode="append_column",
        )  # Now, try to append a column with a different field size

    assert exc_info.type is ValueError
    assert (
        exc_info.value.args[0]
        == "Table sizes don't match. Trying to append column of size 20 to a table of size 10."
    )

    # Finally successfully append a column
    if os.path.exists("/tmp/write_object_append_column_test.lh5"):
        os.remove("/tmp/write_object_append_column_test.lh5")

    tb1 = lh5.Table(col_dict={"dset1": lh5.Array(np.zeros(10))})
    tb2 = lh5.Table(
        col_dict={"dset2": lh5.Array(np.ones(10))}
    )  # different field name, different size
    store = LH5Store()
    store.write_object(
        tb1, "my_table", "/tmp/write_object_append_column_test.lh5", group="my_group"
    )
    store.write_object(
        tb2,
        "my_table",
        "/tmp/write_object_append_column_test.lh5",
        group="my_group",
        wo_mode="append_column",
    )

    # Now, check that the data were appended
    tb_dat, _ = store.read_object(
        "my_group/my_table", "/tmp/write_object_append_column_test.lh5"
    )
    assert isinstance(tb_dat, lgdo.table.Table)
    assert np.array_equal(tb_dat["dset1"].nda, np.zeros(10))
    assert np.array_equal(tb_dat["dset2"].nda, np.ones(10))



import os
import subprocess
from pathlib import Path

config_dir = Path(__file__).parent / "configs"


def test_build_raw_cli(lgnd_test_data):
    subprocess.check_call(["pygama", "build-raw", "--help"])
    subprocess.check_call(
        [
            "pygama",
            "build-raw",
            "--overwrite",
            "--stream-type",
            "ORCA",
            "--out-spec",
            f"{config_dir}/orca-out-spec-cli.json",
            "--max-rows",
            "10",
            "--buffer_size",
            "8192",
            lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca"),
        ]
    )

    assert os.path.exists("/tmp/L200-comm-20220519-phy-geds.lh5")



from __future__ import annotations

import json
import logging
import plistlib

from pygama import lgdo
from pygama.raw.orca.orca_base import OrcaDecoder
from pygama.raw.orca.orca_header import OrcaHeader
from pygama.raw.orca.orca_packet import OrcaPacket
from pygama.raw.raw_buffer import RawBuffer, RawBufferLibrary

log = logging.getLogger(__name__)


class OrcaHeaderDecoder(OrcaDecoder):
    """Decodes ORCA headers."""

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        if header is not None:
            log.warning("unexpected non-None header")
        super().__init__(header=header, **kwargs)
        self.header = OrcaHeader()

    def make_lgdo(self, key: int = None, size: int = None) -> lgdo.Scalar:
        return lgdo.Scalar(value="")

    def buffer_is_full(self, rb: RawBuffer) -> bool:
        return rb.loc > 0

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary = None
    ) -> bool:
        if len(self.header.keys()) != 0:
            log.warning("got an unexpected second header")
        if packet_id != 0:
            log.warning(f"header was packet number {packet_id}")
        n_bytes = packet[1]  # word 1 is the header packet length in bytes
        as_bytes = packet[2:].tobytes()
        self.header.update(plistlib.loads(as_bytes[:n_bytes], fmt=plistlib.FMT_XML))
        if rbl is not None:
            if len(rbl) != 1 or not isinstance(rbl[0].lgdo, lgdo.Scalar):
                raise ValueError(
                    f"invalid RawBufferLibrary object: len(rbl) = {len(rbl)}, type = {type(rbl[0].lgdo)}"
                )
            else:
                rbl[0].lgdo.value = json.dumps(self.header)
                rbl[0].loc = 1
        return True  # header raw buffer is "full"



from __future__ import annotations

from dataclasses import dataclass

import numba
import numpy as np
from numpy import int16, int32, ubyte, uint16, uint32
from numpy.typing import NDArray

from pygama import lgdo

from .base import WaveformCodec

# fmt: off
_radware_sigcompress_mask = uint16([0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023,
                                   2047, 4095, 8191, 16383, 32767, 65535])
# fmt: on


@dataclass(frozen=True)
class RadwareSigcompress(WaveformCodec):
    """`radware-sigcompress` array codec.

    Examples
    --------
    >>> from pygama.lgdo.compression import RadwareSigcompress
    >>> codec = RadwareSigcompress(codec_shift=-32768)
    """

    codec_shift: int = 0
    """Offset added to the input waveform before encoding.

    The `radware-sigcompress` algorithm is limited to encoding of 16-bit
    integer values. In certain cases (notably, with *unsigned* 16-bit integer
    values), shifting incompatible data by a fixed amount circumvents the
    issue.
    """


def encode(
    sig_in: NDArray | lgdo.VectorOfVectors | lgdo.ArrayOfEqualSizedArrays,
    sig_out: NDArray[ubyte]
    | lgdo.VectorOfEncodedVectors
    | lgdo.ArrayOfEncodedEqualSizedArrays = None,
    shift: int32 = 0,
) -> NDArray[ubyte] | lgdo.VectorOfEncodedVectors:
    """Compress digital signal(s) with `radware-sigcompress`.

    Wraps :func:`._radware_sigcompress_encode` and adds support for encoding
    LGDO arrays. Resizes the encoded array to its actual length.

    Note
    ----
    The compression algorithm internally interprets the input waveform values as
    16-bit integers. Make sure that your signal can be safely cast to such a
    numeric type. If not, you may want to apply a `shift` to the waveform.

    Parameters
    ----------
    sig_in
        array(s) holding the input signal(s).
    sig_out
        pre-allocated unsigned 8-bit integer array(s) for the compressed
        signal(s). If not provided, a new one will be allocated.
    shift
        value to be added to `sig_in` before compression.

    Returns
    -------
    sig_out
        given pre-allocated `sig_out` structure or new structure of unsigned
        8-bit integers.

    See Also
    --------
    ._radware_sigcompress_encode
    """
    # the encoded signal is an array of bytes -> twice as long as a uint16
    # array
    max_out_len = 2 * len(sig_in)
    if isinstance(sig_in, np.ndarray) and sig_in.ndim == 1:
        if len(sig_in) == 0:
            return sig_in

        if not sig_out:
            # pre-allocate ubyte (uint8) array
            sig_out = np.empty(max_out_len, dtype=ubyte)

        if sig_out.dtype != ubyte:
            raise ValueError("sig_out must be of type ubyte")

        outlen = _radware_sigcompress_encode(sig_in, sig_out, shift=shift)

        # resize (down) the encoded signal to its actual length
        # TODO: really even if user supplied? Maybe not
        if outlen < max_out_len:
            sig_out.resize(outlen, refcheck=True)

    elif isinstance(sig_in, lgdo.ArrayOfEqualSizedArrays):
        if not sig_out:
            # pre-allocate output structure
            # use maximum length possible
            sig_out = lgdo.ArrayOfEncodedEqualSizedArrays(
                encoded_data=lgdo.VectorOfVectors(
                    shape_guess=(len(sig_in), 2 * sig_in.nda.shape[1]), dtype=ubyte
                ),
                decoded_size=sig_in.nda.shape[1],
            )
        elif not isinstance(sig_out, lgdo.ArrayOfEncodedEqualSizedArrays):
            raise ValueError("sig_out must be an ArrayOfEncodedEqualSizedArrays")

        # use unsafe set_vector to fill pre-allocated memory
        for i, wf in enumerate(sig_in):
            sig_out.encoded_data._set_vector_unsafe(i, encode(wf, shift=shift))

        # resize down flattened data array
        # TODO: really even if user supplied? Maybe not
        sig_out.resize(len(sig_in))

    elif isinstance(sig_in, lgdo.VectorOfVectors):
        if not sig_out:
            max_out_len = 2 * len(sig_in.flattened_data) / len(sig_in.cumulative_length)
            sig_out = lgdo.VectorOfEncodedVectors(
                encoded_data=lgdo.VectorOfVectors(
                    shape_guess=(len(sig_in), max_out_len), dtype=ubyte
                ),
            )
        elif not isinstance(sig_out, lgdo.VectorOfEncodedVectors):
            raise ValueError("sig_out must be a VectorOfEncodedVectors")

        # use unsafe set_vector to fill pre-allocated memory
        # should be fast enough
        for i, wf in enumerate(sig_in):
            sig_out.encoded_data._set_vector_unsafe(i, encode(wf, shift=shift))
            sig_out.decoded_size[i] = len(wf)

    else:
        raise ValueError(f"unsupported input signal type ({type(sig_in)})")

        # resize down flattened data array
        # TODO: really even if user supplied? Maybe not
        sig_out.resize(len(sig_in))

    return sig_out


def decode(
    sig_in: NDArray[ubyte]
    | lgdo.VectorOfEncodedVectors
    | lgdo.ArrayOfEncodedEqualSizedArrays,
    sig_out: NDArray | lgdo.VectorOfVectors | lgdo.ArrayOfEqualSizedArrays = None,
    shift: int32 = 0,
) -> NDArray | lgdo.VectorOfVectors | lgdo.ArrayOfEqualSizedArrays:
    """Decompress digital signal(s) with `radware-sigcompress`.

    Wraps :func:`._radware_sigcompress_decode` and adds support for decoding
    LGDOs. Resizes the decoded signals to their actual length.

    Parameters
    ----------
    sig_in
        array(s) holding the input, compressed signal(s).
    sig_out
        pre-allocated array(s) for the decompressed signal(s).  If not
        provided, will allocate a 32-bit integer array(s) structure.
    shift
        the value the original signal(s) was shifted before compression.  The
        value is *subtracted* from samples in `sig_out` right after decoding.

    Returns
    -------
    sig_out
        given pre-allocated structure or new structure of 32-bit integers.

    See Also
    --------
    ._radware_sigcompress_decode
    """
    if isinstance(sig_in, np.ndarray) and sig_in.ndim == 1 and sig_in.dtype == ubyte:
        if len(sig_in) == 0:
            return sig_in

        siglen = _get_hton_u16(sig_in, 0)
        if not sig_out:
            # pre-allocate memory, use safe int32
            sig_out = np.empty(siglen, dtype="int32")
        elif len(sig_out) < siglen:
            # TODO: really even if user supplied? Maybe not
            sig_out.resize(siglen, refcheck=False)

        _radware_sigcompress_decode(sig_in, sig_out, shift=shift)

    elif isinstance(sig_in, lgdo.ArrayOfEncodedEqualSizedArrays):
        if not sig_out:
            # pre-allocate output structure
            sig_out = lgdo.ArrayOfEqualSizedArrays(
                dims=(1, 1),
                shape=(len(sig_in), sig_in.decoded_size.value),
                dtype="int32",
            )

        elif not isinstance(sig_out, lgdo.ArrayOfEqualSizedArrays):
            raise ValueError("sig_out must be an ArrayOfEqualSizedArrays")

        for i, wf in enumerate(sig_in):
            sig_out[i] = decode(wf, shift=shift)

    elif isinstance(sig_in, lgdo.VectorOfEncodedVectors):
        if not sig_out:
            # pre-allocate output structure
            sig_out = lgdo.VectorOfVectors(
                cumulative_length=np.cumsum(sig_in.decoded_size), dtype="int32"
            )

        elif not isinstance(sig_out, lgdo.VectorOfVectors):
            raise ValueError("sig_out must be a VectorOfVectors")

        for i, wf in enumerate(sig_in):
            sig_out[i] = decode(wf[0], shift=shift)

    else:
        raise ValueError(f"unsupported input signal type ({type(sig_in)})")

    return sig_out


@numba.jit(nopython=True)
def _set_hton_u16(a: NDArray[ubyte], i: int, x: int) -> int:
    """Store an unsigned 16-bit integer value in an array of unsigned 8-bit integers.

    The first two most significant bytes from `x` are stored contiguously in
    `a` with big-endian order.
    """
    x_u16 = uint16(x)
    i_1 = i * 2
    i_2 = i_1 + 1
    a[i_1] = ubyte(x_u16 >> 8)
    a[i_2] = ubyte(x_u16 >> 0)
    return x


@numba.jit(nopython=True)
def _get_hton_u16(a: NDArray[ubyte], i: int) -> uint16:
    """Read unsigned 16-bit integer values from an array of unsigned 8-bit integers.

    The first two most significant bytes of the values must be stored
    contiguously in `a` with big-endian order.
    """
    i_1 = i * 2
    i_2 = i_1 + 1
    return uint16(a[i_1] << 8 | a[i_2])


@numba.jit("uint16(uint32)", nopython=True)
def _get_high_u16(x: uint32) -> uint16:
    return uint16(x >> 16)


@numba.jit("uint32(uint32, uint16)", nopython=True)
def _set_high_u16(x: uint32, y: uint16) -> uint32:
    return uint32(x & 0x0000FFFF | (y << 16))


@numba.jit("uint16(uint32)", nopython=True)
def _get_low_u16(x: uint32) -> uint16:
    return uint16(x >> 0)


@numba.jit("uint32(uint32, uint16)", nopython=True)
def _set_low_u16(x: uint32, y: uint16) -> uint32:
    return uint32(x & 0xFFFF0000 | (y << 0))


@numba.jit(nopython=True)
def _radware_sigcompress_encode(
    sig_in: NDArray,
    sig_out: NDArray[ubyte],
    shift: int32,
    _mask: NDArray[uint16] = _radware_sigcompress_mask,
) -> int32:
    """Compress a digital signal.

    Shifts the signal values by ``+shift`` and internally interprets the result
    as :any:`numpy.int16`. Shifted signals must be therefore representable as
    :any:`numpy.int16`, for lossless compression.

    Note
    ----
    The algorithm also computes the first derivative of the input signal, which
    cannot always be represented as a 16-bit integer. In such cases, overflows
    occur, but they seem to be innocuous.

    Almost literal translations of ``compress_signal()`` from the
    `radware-sigcompress` v1.0 C-code by David Radford [1]_. Summary of
    changes:

    - Shift the input signal by `shift` before encoding.
    - Store encoded, :class:`numpy.uint16` signal as an array of bytes
      (:class:`numpy.ubyte`), in big-endian ordering.
    - Declare mask globally to avoid extra memory allocation.
    - Apply just-in-time compilation with Numba.
    - Add a couple of missing array boundary checks.

    .. [1] `radware-sigcompress source code
       <https://legend-exp.github.io/legend-data-format-specs/dev/data_compression/#radware-sigcompress-1>`_.
       released under MIT license `[Copyright (c) 2018, David C. Radford
       <radforddc@ornl.gov>]`.

    Parameters
    ----------
    sig_in
        array of integers holding the input signal. In the original C code,
        an array of 16-bit integers was expected.
    sig_out
        pre-allocated array for the unsigned 8-bit encoded signal. In the
        original C code, an array of unsigned 16-bit integers was expected.

    Returns
    -------
    length
        number of bytes in the encoded signal
    """
    mask = _mask

    i = j = max1 = max2 = min1 = min2 = ds = int16(0)
    nb1 = nb2 = iso = nw = bp = dd1 = dd2 = int16(0)
    dd = uint32(0)

    _set_hton_u16(sig_out, iso, sig_in.size)

    iso += 1
    while j < sig_in.size:  # j = starting index of section of signal
        # find optimal method and length for compression
        # of next section of signal
        si_j = int16(sig_in[j] + shift)
        max1 = min1 = si_j
        max2 = int32(-16000)
        min2 = int32(16000)
        nb1 = nb2 = 2
        nw = 1
        i = j + 1
        # FIXME: 48 could be tuned better?
        while (i < sig_in.size) and (i < j + 48):
            si_i = int16(sig_in[i] + shift)
            si_im1 = int16(sig_in[i - 1] + shift)
            if max1 < si_i:
                max1 = si_i
            if min1 > si_i:
                min1 = si_i
            ds = si_i - si_im1
            if max2 < ds:
                max2 = ds
            if min2 > ds:
                min2 = ds
            nw += 1
            i += 1
        if max1 - min1 <= max2 - min2:  # use absolute values
            nb2 = 99
            while (max1 - min1) > mask[nb1]:
                nb1 += 1
            while (i < sig_in.size) and (
                i < j + 128
            ):  # FIXME: 128 could be tuned better?
                si_i = int16(sig_in[i] + shift)
                if max1 < si_i:
                    max1 = si_i
                dd1 = max1 - min1
                if min1 > si_i:
                    dd1 = max1 - si_i
                if dd1 > mask[nb1]:
                    break
                if min1 > si_i:
                    min1 = si_i
                nw += 1
                i += 1
        else:  # use difference values
            nb1 = 99
            while max2 - min2 > mask[nb2]:
                nb2 += 1
            while (i < sig_in.size) and (
                i < j + 128
            ):  # FIXME: 128 could be tuned better?
                si_i = int16(sig_in[i] + shift)
                si_im1 = int16(sig_in[i - 1] + shift)
                ds = si_i - si_im1
                if max2 < ds:
                    max2 = ds
                dd2 = max2 - min2
                if min2 > ds:
                    dd2 = max2 - ds
                if dd2 > mask[nb2]:
                    break
                if min2 > ds:
                    min2 = ds
                nw += 1
                i += 1

        if bp > 0:
            iso += 1
        # do actual compression
        _set_hton_u16(sig_out, iso, nw)
        iso += 1
        bp = 0
        if nb1 <= nb2:
            # encode absolute values
            _set_hton_u16(sig_out, iso, nb1)
            iso += 1
            _set_hton_u16(sig_out, iso, uint16(min1))
            iso += 1

            i = iso
            while i <= (iso + nw * nb1 / 16):
                _set_hton_u16(sig_out, i, 0)
                i += 1

            i = j
            while i < j + nw:
                dd = int16(sig_in[i] + shift) - min1  # value to encode
                dd = dd << (32 - bp - nb1)
                _set_hton_u16(
                    sig_out, iso, _get_hton_u16(sig_out, iso) | _get_high_u16(dd)
                )
                bp += nb1
                if bp > 15:
                    iso += 1
                    _set_hton_u16(sig_out, iso, _get_low_u16(dd))
                    bp -= 16
                i += 1

        else:
            # encode derivative / difference values
            _set_hton_u16(sig_out, iso, nb2 + 32)  # bits used for encoding, plus flag
            iso += 1
            _set_hton_u16(sig_out, iso, int16(si_j))  # starting signal value
            iso += 1
            _set_hton_u16(sig_out, iso, int16(min2))  # min value used for encoding
            iso += 1

            i = iso
            while i <= iso + nw * nb2 / 16:
                _set_hton_u16(sig_out, i, 0)
                i += 1

            i = j + 1
            while i < j + nw:
                si_i = int16(sig_in[i] + shift)
                si_im1 = int16(sig_in[i - 1] + shift)
                dd = si_i - si_im1 - min2
                dd = dd << (32 - bp - nb2)
                _set_hton_u16(
                    sig_out, iso, _get_hton_u16(sig_out, iso) | _get_high_u16(dd)
                )
                bp += nb2
                if bp > 15:
                    iso += 1
                    _set_hton_u16(sig_out, iso, _get_low_u16(dd))
                    bp -= 16
                i += 1
        j += nw

    if bp > 0:
        iso += 1

    if iso % 2 > 0:
        iso += 1

    return 2 * iso  # number of bytes in compressed signal data


@numba.jit(nopython=True)
def _radware_sigcompress_decode(
    sig_in: NDArray[ubyte],
    sig_out: NDArray,
    shift: int32,
    _mask: NDArray[uint16] = _radware_sigcompress_mask,
) -> int32:
    """Deompress a digital signal.

    After decoding, the signal values are shifted by ``-shift`` to restore the
    original waveform. The dtype of `sig_out` must be large enough to contain it.

    Almost literal translations of ``decompress_signal()`` from the
    `radware-sigcompress` v1.0 C-code by David Radford [1]_. See
    :func:`._radware_sigcompress_encode` for a list of changes to the original
    algorithm.

    Parameters
    ----------
    sig_in
        array holding the input, compressed signal. In the original code, an
        array of 16-bit unsigned integers was expected.
    sig_out
        pre-allocated array for the decompressed signal. In the original code,
        an array of 16-bit integers was expected.

    Returns
    -------
    length
        length of output, decompressed signal.
    """
    mask = _mask

    i = j = min_val = nb = isi = iso = nw = bp = int16(0)
    dd = uint32(0)

    sig_len_in = int(sig_in.size / 2)
    siglen = int16(_get_hton_u16(sig_in, isi))  # signal length
    isi += 1

    while (isi < sig_len_in) and (iso < siglen):
        if bp > 0:
            isi += 1
        bp = 0  # bit pointer
        nw = _get_hton_u16(sig_in, isi)  # number of samples encoded in this chunk
        isi += 1
        nb = _get_hton_u16(sig_in, isi)  # number of bits used in compression
        isi += 1

        if nb < 32:
            # decode absolute values
            min_val = int16(_get_hton_u16(sig_in, isi))  # min value used for encoding
            isi += 1
            dd = _set_low_u16(dd, _get_hton_u16(sig_in, isi))
            i = 0
            while (i < nw) and (iso < siglen):
                if (bp + nb) > 15:
                    bp -= 16
                    dd = _set_high_u16(dd, _get_hton_u16(sig_in, isi))
                    isi += 1
                    if isi < sig_len_in:
                        dd = _set_low_u16(dd, _get_hton_u16(sig_in, isi))
                    dd = dd << (bp + nb)
                else:
                    dd = dd << nb
                sig_out[iso] = (_get_high_u16(dd) & mask[nb]) + min_val - shift
                iso += 1
                bp += nb
                i += 1
        else:
            nb -= 32
            #  decode derivative / difference values
            sig_out[iso] = (
                int16(_get_hton_u16(sig_in, isi)) - shift
            )  # starting signal value
            iso += 1
            isi += 1
            min_val = int16(_get_hton_u16(sig_in, isi))  # min value used for encoding
            isi += 1
            if isi < sig_len_in:
                dd = _set_low_u16(dd, _get_hton_u16(sig_in, isi))

            i = 1
            while (i < nw) and (iso < siglen):
                if (bp + nb) > 15:
                    bp -= 16
                    dd = _set_high_u16(dd, _get_hton_u16(sig_in, isi))
                    isi += 1
                    if isi < sig_len_in:
                        dd = _set_low_u16(dd, _get_hton_u16(sig_in, isi))
                    dd = dd << (bp + nb)
                else:
                    dd = dd << nb
                sig_out[iso] = (
                    int16(
                        (_get_high_u16(dd) & mask[nb])
                        + min_val
                        + sig_out[iso - 1]
                        + shift
                    )
                    - shift
                )
                iso += 1
                bp += nb
                i += 1
        j += nw

    if siglen != iso:
        raise RuntimeError("failure: unexpected signal length after decompression")

    return siglen  # number of shorts in decompressed signal data



import os

import numpy as np
import pytest

import pygama.lgdo.lgdo_utils as lgdo_utils


def test_get_element_type():
    objs = [
        ("hi", "string"),
        (True, "bool"),
        (np.void(0), "blob"),
        (int(0), "real"),
        (np.uint8(0), "real"),
        (float(0), "real"),
        (1 + 1j, "complex"),
        (b"hi", "string"),
        (np.array(["hi"]), "string"),
    ]

    for obj, name in objs:
        get_name = lgdo_utils.get_element_type(obj)
        assert get_name == name


def test_parse_datatype():
    datatypes = [
        ("real", ("scalar", None, "real")),
        ("array<1>{bool}", ("array", (1,), "bool")),
        ("fixedsizearray<2>{real}", ("fixedsizearray", (2,), "real")),
        (
            "arrayofequalsizedarrays<3,4>{complex}",
            ("arrayofequalsizedarrays", (3, 4), "complex"),
        ),
        ("array<1>{array<1>{blob}}", ("array", (1,), "array<1>{blob}")),
        (
            "struct{field1,field2,fieldn}",
            ("struct", None, ["field1", "field2", "fieldn"]),
        ),
        ("table{col1,col2,coln}", ("table", None, ["col1", "col2", "coln"])),
    ]

    for string, dt_tuple in datatypes:
        pd_dt_tuple = lgdo_utils.parse_datatype(string)
        assert pd_dt_tuple == dt_tuple


def test_expand_vars():
    # Check env variable expansion
    os.environ["PYGAMATESTBASEDIR"] = "a_random_string"
    assert lgdo_utils.expand_vars("$PYGAMATESTBASEDIR/blah") == "a_random_string/blah"

    # Check user variable expansion
    assert (
        lgdo_utils.expand_vars(
            "$PYGAMATESTBASEDIR2/blah",
            substitute={"PYGAMATESTBASEDIR2": "a_random_string"},
        )
        == "a_random_string/blah"
    )


def test_expand_path(lgnd_test_data):
    files = [
        lgnd_test_data.get_path(
            "lh5/prod-ref-l200/generated/tier/dsp/cal/p01/r014/l60-p01-r014-cal-20220716T104550Z-tier_dsp.lh5"
        ),
        lgnd_test_data.get_path(
            "lh5/prod-ref-l200/generated/tier/dsp/cal/p01/r014/l60-p01-r014-cal-20220716T105236Z-tier_dsp.lh5"
        ),
    ]
    base_dir = os.path.dirname(files[0])

    assert lgdo_utils.expand_path(f"{base_dir}/*20220716T104550Z*") == files[0]

    # Should fail if file not found
    with pytest.raises(FileNotFoundError):
        lgdo_utils.expand_path(f"{base_dir}/not_a_real_file.lh5")

    # Should fail if multiple files found
    with pytest.raises(FileNotFoundError):
        lgdo_utils.expand_path(f"{base_dir}/*.lh5")

    # Check if it finds a list of files correctly
    assert sorted(lgdo_utils.expand_path(f"{base_dir}/*.lh5", list=True)) == sorted(
        files
    )



import inspect

import numpy as np
import pytest

from pygama.dsp.errors import DSPFatal
from pygama.dsp.processors import get_multi_local_extrema

#                .
#               . .
#              .   .
#             .     .
#            .       .
#       .   .         .   .
#      . . .           . . .
#     .   .             .   .
#    .                       .
#   .                         .
# ..                           ..
# 0123456789012345678901234567890
#           1         2         3
wf = np.array([0, 0, 1, 2, 3, 4, 5, 4, 3, 4, 5, 6, 7, 8, 9, 10,
               9, 8, 7, 6, 5, 4, 3, 4, 5, 4, 3, 2, 1, 0, 0])  # fmt: skip


# delta_min=delta_max tests (all searches should behave identical)


# L->R
def test_get_multi_local_extrema_ltor(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        3,
        0,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([np.nan, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1
    assert n_min_out[0] == 0


# L<-R
def test_get_multi_local_extrema_rtol(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        3,
        1,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([np.nan, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1
    assert n_min_out[0] == 0


# L<-c->R
def test_get_multi_local_extrema_both_cons(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        3,
        2,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([np.nan, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1


# L<-a->R
def test_get_multi_local_extrema_both_agro(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        3,
        3,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([np.nan, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1


# delta_min != delta_max (result should be search direction dependent)


# L->R
def test_get_multi_local_extrema_ltor_asym(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        1,
        0,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, 24, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([22, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 2
    assert n_min_out[0] == 1


# L<-R
def test_get_multi_local_extrema_rtol_asym(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        1,
        1,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, 6, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([8, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 2
    assert n_min_out[0] == 1


# L<-c->R
def test_get_multi_local_extrema_both_cons_asym(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        1,
        2,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([np.nan, np.nan, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1
    assert n_min_out[0] == 0


# L<-a->R
def test_get_multi_local_extrema_both_agro_asym(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        1,
        3,
        0,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([6, 15, 24]), equal_nan=True)
    assert np.array_equal(min_out, np.array([8, 22, np.nan]), equal_nan=True)
    assert n_max_out[0] == 3
    assert n_min_out[0] == 2


# Absolute threshold test with aggressive both direction search
def test_get_multi_local_extrema_both_agro_asym_abs(compare_numba_vs_python):
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    compare_numba_vs_python(
        get_multi_local_extrema,
        wf,
        3,
        1,
        3,
        8,
        20,
        max_out,
        min_out,
        n_max_out,
        n_min_out,
    )

    assert np.array_equal(max_out, np.array([15, np.nan, np.nan]), equal_nan=True)
    assert np.array_equal(min_out, np.array([8, 22, np.nan]), equal_nan=True)
    assert n_max_out[0] == 1
    assert n_min_out[0] == 2


# And now test if everything works well when vectorized


# return nan tests
def test_get_multi_local_extrema_return_on_nan(compare_numba_vs_python):
    wf = np.ones(20)
    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    wf[4] = np.nan
    assert np.isnan(
        compare_numba_vs_python(
            get_multi_local_extrema,
            wf,
            3,
            1,
            3,
            8,
            20,
            max_out,
            min_out,
            n_max_out,
            n_min_out,
        )
    )

    wf[4] = 3
    assert np.isnan(
        compare_numba_vs_python(
            get_multi_local_extrema,
            wf,
            np.nan,
            3,
            3,
            8,
            20,
            max_out,
            min_out,
            n_max_out,
            n_min_out,
        )
    )
    assert np.isnan(
        compare_numba_vs_python(
            get_multi_local_extrema,
            wf,
            np.nan,
            3,
            np.nan,
            8,
            20,
            max_out,
            min_out,
            n_max_out,
            n_min_out,
        )
    )
    assert np.isnan(
        compare_numba_vs_python(
            get_multi_local_extrema,
            wf,
            np.nan,
            np.nan,
            3,
            8,
            20,
            max_out,
            min_out,
            n_max_out,
            n_min_out,
        )
    )


# DSP Fatal test
def test_get_multi_local_extrema_dsp_fatal():
    wf = np.ones(20)
    max_out = np.zeros(len(wf) + 1)
    max_out[:] = np.nan
    min_out = np.zeros(len(wf) + 1)
    min_out[:] = np.nan
    n_min_out = np.zeros(1)
    n_max_out = np.zeros(1)

    with pytest.raises(DSPFatal):
        get_multi_local_extrema(
            wf, 3, 1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )
    with pytest.raises(DSPFatal):
        inspect.unwrap(get_multi_local_extrema)(
            wf, 3, 1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    max_out = np.zeros(3)
    max_out[:] = np.nan
    min_out = np.zeros(3)
    min_out[:] = np.nan

    with pytest.raises(DSPFatal):
        get_multi_local_extrema(
            wf, -3, 1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    with pytest.raises(DSPFatal):
        inspect.unwrap(get_multi_local_extrema)(
            wf, -3, 1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    with pytest.raises(DSPFatal):
        get_multi_local_extrema(
            wf, 3, -1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    with pytest.raises(DSPFatal):
        inspect.unwrap(get_multi_local_extrema)(
            wf, 3, -1, 3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    with pytest.raises(DSPFatal):
        get_multi_local_extrema(
            wf, 3, 1, -3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )

    with pytest.raises(DSPFatal):
        inspect.unwrap(get_multi_local_extrema)(
            wf, 3, 1, -3, 8, 20, max_out, min_out, n_max_out, n_min_out
        )



import pytest

from pygama.raw.compass.compass_header_decoder import CompassHeaderDecoder
from pygama.raw.compass.compass_streamer import CompassStreamer


@pytest.fixture(scope="module")
def packet(lgnd_test_data):
    streamer = CompassStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"), buffer_size=6
    )
    init_rbytes = streamer.n_bytes_read
    pkt = streamer.load_packet()
    assert pkt is not None  # load was successful
    assert (
        streamer.n_bytes_read
        == init_rbytes + 25 + 2 * streamer.header["boards"]["0"]["wf_len"].value
    )
    streamer.close_stream()
    return pkt


@pytest.fixture(scope="module")
def compass_config(lgnd_test_data):
    decoder = CompassHeaderDecoder()
    test_file = lgnd_test_data.get_path("compass/compass_test_data.BIN")
    in_stream = open(test_file, "rb")
    compass_config_file = lgnd_test_data.get_path(
        "compass/compass_test_data_settings.xml"
    )
    decoder.decode_header(in_stream, compass_config_file)
    in_stream.close()
    return decoder.config


@pytest.fixture(scope="module")
def compass_config_no_settings(lgnd_test_data):
    decoder = CompassHeaderDecoder()
    test_file = lgnd_test_data.get_path("compass/compass_test_data.BIN")
    in_stream = open(test_file, "rb")
    compass_config_file = None
    wf_len = 1000
    decoder.decode_header(in_stream, compass_config_file, wf_len)
    in_stream.close()
    return decoder.config



from __future__ import annotations

import logging

import fcutils
import numpy as np

from pygama import lgdo
from pygama.raw.data_decoder import DataDecoder
from pygama.raw.raw_buffer import RawBuffer

log = logging.getLogger(__name__)


class FCConfigDecoder(DataDecoder):
    """Decode FlashCam config data.

    Note
    ----
    Derives from :class:`~.raw.data_decoder.DataDecoder` in anticipation of
    possible future functionality. Currently the base class interface is not
    used.

    Example
    -------
    >>> import fcutils
    >>> from pygama.raw.fc.fc_config_decoder import FCConfigDecoder
    >>> fc = fcutils.fcio('file.fcio')
    >>> decoder = FCConfigDecoder()
    >>> config = decoder.decode_config(fc)
    >>> type(config)
    pygama.lgdo.struct.Struct
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.config = lgdo.Struct()

    def decode_config(self, fcio: fcutils.fcio) -> lgdo.Struct:
        config_names = [
            "nsamples",  # samples per channel
            "nadcs",  # number of adc channels
            "ntriggers",  # number of triggertraces
            "telid",  # id of telescope
            "adcbits",  # bit range of the adc channels
            "sumlength",  # length of the fpga integrator
            "blprecision",  # precision of the fpga baseline
            "mastercards",  # number of attached mastercards
            "triggercards",  # number of attached triggercards
            "adccards",  # number of attached fadccards
            "gps",  # gps mode (0: not used, 1: external pps and 10MHz)
        ]
        for name in config_names:
            if name in self.config:
                log.warning(f"{name} already in self.config. skipping...")
                continue
            value = np.int32(getattr(fcio, name))  # all config fields are int32
            self.config.add_field(name, lgdo.Scalar(value))
        return self.config

    def make_lgdo(self, key: int = None, size: int = None) -> lgdo.Struct:
        return self.config

    def buffer_is_full(self, rb: RawBuffer) -> bool:
        return rb.loc > 0
